import os
from requests import HTTPError
from client.user import User
from pipeline import msg

class NewtClient(User):
    """
    Client class to handle Newt API calls

    Attributes
    ----------
    authentication : str
        authentication token
    scratch_dir : str
        path to a user's scratch directory on NERSC system
    home_dir : str
        path to a user's scratch directory on NERSC system
    """

    BASE_URL = "https://newt.nersc.gov/newt"
    systems = ('edison', 'cori')

    def __init__(self):
        super(NewtClient, self).__init__()
        self.authentication = None  # NewtClient.login(self)
        self.scratch_dir = None
        self.home_dir = None

    def __del__(self):
        if self.logged_in:
            self.logout()
        try:
            self.session.close()
        except TypeError:
            pass

    def login(self, username, password):
        """
        Login to NEWT

        Parameters
        ----------
        username : str
        password : str
        """

        credentials = {"username": username,
                       "password": password}

        response = self.post(self.BASE_URL + "/auth", data=credentials)

        if response.json()['auth']:
            self.authentication = response
            return super(NewtClient, self).login(username)
        else:
            self.authentication = None
            raise NEWTError('Bad Authentication: Unable to log in')

    def logout(self):
        """
        Send REST call to logout from NEWT
        """
        self.authentication = self.post(self.BASE_URL + '/logout')
        User.logout(self)

    @classmethod
    def check_system(cls, system):
        """
        Check if the given system is in the class's allowed systems list.
        Raise an error if not

        Parameters
        ----------
        system : str
            NERSC system name
        """

        if system not in cls.systems:
            raise NEWTError('%s is not a NERSC system' % system)

    def get_system_status(self, system='all'):
        """
        Get system status via NEWT API status call

        Parameters
        ----------
        system : str
            NERSC system name ('all' for all systems)

        Returns
        -------
        json
            status response
        """

        self.check_login()
        if system is 'all':
            system = ''
        else:
            self.check_system(system)

        r = self.post(self.BASE_URL + '/status/' + system)
        return self.check_response(r)

    def get_dir_contents(self, path, system):
        """
        Get contents of a directory on a NERSC system

        Parameters
        ----------
        path : str
            path of directory
        system : str
            NERSC system name

        Returns
        -------
        json
            response of directory contents
        """

        self.check_login()
        self.check_system(system)

        r = self.get(self.BASE_URL + '/file/' + system + path)

        return self.check_response(r)

    def get_file_size(self, path, system):
        """
        Get the size of a file in bytes

        Parameters
        ----------
        path : str
            path to file

        Returns
        -------
        int
            file size
        """

        info = self.get_dir_contents(path, system)
        if len(info) > 1:
            raise NEWTError('{} is a directory'.format(path))

        return int(info[0]['size'])

    def get_scratch_dir(self, system):
        """
        Get a users scratch directory on a NERSC platform

        Parameters
        ----------
        system : str, optional
            NERSC system name

        Returns
        -------
        str
            path of user's scratch if found. None, if not found
        """

        self.check_login()
        scratch_dir = None
        if system == 'edison':
            for i in range(1, 10):
                root_folder = ('/scratch' + str(i) + '/scratchdirs/' +
                               self.username)
                if self._check_scratch_path(root_folder, system):
                    scratch_dir = root_folder
        elif system == 'cori':
            root_folder = '/global/cscratch1/sd/' + self.username
            if self._check_scratch_path(root_folder, system):
                scratch_dir = root_folder

        if scratch_dir is None:
            # raise NEWTError('User %s scratch directory not found'
            #                  % self.username)
            return '/'
        return scratch_dir

    def set_scratch_dir(self, system):
        """
        Get a users scratch directory on a NERSC platform and set it to the instance variable

        Parameters
        ----------
        system : str, optional
            NERSC system name
        """
        self.scratch_dir = self.get_scratch_dir(system)

    def _check_scratch_path(self, path, system):
        """
        Checks to see if the path is accessible by NEWT

        Parameters
        ----------
        path : str
        system : str
            NERSC system
        """

        try:
            contents = self.get_dir_contents(path, system)
            for item in contents:
                if 'user' in item.keys():
                    if item['user'] == self.username:
                        return True
        except HTTPError:
            return False

    def get_home_dir(self, system):
        """
        Get a user's home directory on a NERSC system

        Parameters
        ----------
        system : str
            NERSC system name

        Returns
        -------
        str
            path of users home directory if found. None, if not found
        """

        self.check_login()
        home_dir = None
        try:
            root_home = ('/global/homes/' + self.username[0] + '/' +
                         self.username)
            contents = self.get_dir_contents(root_home, system)
            for item in contents:
                    if 'user' in item.keys():
                        if item['user'] == self.username:
                            home_dir = root_home
                            break
        except:
            pass

        return home_dir

    def set_home_dir(self, system):
        """
        Get a users home directory on a NERSC platform and set it to the instance variable

        Parameters
        ----------
        system : str, optional
            NERSC system name
        """

        self.home_dir = self.get_home_dir(system)

    def execute_command(self, command, system, bin_path='/bin'):
        """
        Execute a system command on a NERSC system

        Parameters
        ----------
        command : str
            command including any arguments
        system : str
            NERSC system name
        bin_path : str
            path to binary
        """

        self.check_login()
        self.check_system(system)

        r = self.post(self.BASE_URL + '/command/' + system,
                      data={'executable': bin_path + '/' + command})

        return self.check_response(r)

    def delete_file(self, fpath, system):
        """
        Delete a specific file on a NERSC system, 'rm -rf' command

        Parameters
        ----------
        fpath : str
            path to file
        system : str
            NERSC system name
        """

        command = 'rm -rf ' + fpath
        return self.execute_command(command, system)

    def move_file(self, fpath, dst_path, system):
        """
        Move a specified file on a NERSC system. 'mv' command

        Parameters
        ----------
        fpath : str
            path to file
        dst_path : str
            path to new directory
        system: str
            NERSC system name
        """

        command = 'mv ' + fpath + ' ' + dst_path
        return self.execute_command(command, system)

    def copy_file(self, fpath, dst_path, system):
        """
        Copy a specified file to a new location on a NERSC system.
        'cp -ar' command

        Parameters
        ----------
        fpath : str
            path to file
        dst_path : str
            path to copy into
        system : str
            NERSC system name
        """
        command = 'cp -ar ' + fpath + ' ' + dst_path
        return self.execute_command(command, system)

    def rsync(self, fpath, dst_path, system, *args):
        """
        Use rsync locally on a nersc system

        Parameters
        ----------
        fpath : str
            source path
        dst_path : str
            dest path
        system : str
            nersc system
        args : str
            lags for rsync (ie '-p'
        """
        command = 'rsync '
        for arg in args:
            command += arg + ' '
        command += fpath + ' ' + dst_path

        return self.execute_command(command, system, bin_path='/usr/bin')

    def upload_file(self, fpath, dpath, system):
        """
        Upload a file (<100 MB or it will not work)

        Parameters
        ----------
        fpath : str
            Local path to file
        dpath : str
            Destination path
        system : str
            NERSC system
        """

        self.check_login()
        self.check_system(system)
        fname = os.path.split(fpath)[-1]
        with open(fpath, 'rb') as f:
            file = {'file': (fpath, f), 'file_name': fname}
            response = self.post(self.BASE_URL + '/file/' + system + dpath, files=file)

        return response

    def download_file(self, path, system, save_path=None):
        """
        Download a file on a NERSC system. (<100 MB or it will not work)

        Parameters
        ----------
        path : str
            file path on NERSC
        system : str
            NERSC system
        save_path : str
            path and name to save file locally. If None name on NERSC is used and save in home directory
        :return:
        """

        self.check_login()

        if save_path is None:
            save_path = os.path.join(os.path.expanduser('~'), os.path.split(path)[-1])

        r = self.get(self.BASE_URL + '/file/' + system + '/' +  path + '?view=read', stream=True)

        with open(save_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=64*1024):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    f.flush()
        r.close()

        return

    def download_file_generator(self, path, system, save_path=None, chunk_size=64*1024):
        """
        Download a dataset as a generator (yields the fraction downloaded)
        Useful to know the status of a download (for gui purposes)

        Parameters
        ----------
        path : str
            file path on NERSC
        system : str
            NERSC system
        save_path : str
            path and name to save file locally. If None name on NERSC is used and save in home directory
        chunk_size: in, optional
            Chunks size in bytes

        Yields
        ------
        float
            percent downloaded
        """
        self.check_login()

        if save_path is None:
            save_path = os.path.join(os.path.expanduser('~'), os.path.split(path)[-1])

        file_size = float(self.get_file_size(path, system))
        r = self.get(self.BASE_URL + '/file/' + system + '/' + path + '?view=read', stream=True)

        with open(save_path, 'wb') as f:
            downloaded = 0.0
            for chunk in r.iter_content(chunk_size=chunk_size):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    f.flush()
                    downloaded += len(chunk)/file_size
                    yield downloaded
        r.close()


class NEWTError(Exception):
    """Raised when a parameter that will be used for the NEWT API is invalid"""
    pass



from .. import base


import matplotlib.pyplot as plt

import numpy as np

import xasloader
from xicam.widgets.NDTimelinePlotWidget import XASTimelineWidget
from xicam.widgets import fitting
from PySide.QtGui import *

def runtest():
    EZTest.bottomwidget.clear()

def openfiles(filepaths):
    for path in filepaths:
        spectra = xasloader.open(path)
        spectra.treat()

        for t,scan in enumerate(spectra.scans):
            EZTest.plot((t,(np.array(scan.Energy), np.array(scan.I_norm))))


XASplot = XASTimelineWidget()

EZTest=base.EZplugin(name='XAS',
                     toolbuttons=[('xicam/gui/icons_34.png',runtest)],
                     parameters=[{'name':'Pre-edge Min','value':10,'type':'int'},
                                 {'name':'Pre-edge Max','value':30,'type':'int'},
                                 fitting.FitParameter(XASplot.stackplot.plotWidget)],
                     openfileshandler=openfiles,
                     centerwidget=None,
                     bottomwidget=XASplot)


import json
from PySide.QtUiTools import QUiLoader
from PySide import QtGui
from PySide import QtCore
import yaml
from modpkgs.collectionsmod import UnsortableOrderedDict
import ui
import featuremanager
import display
import customwidgets


class mainwindow():
    def __init__(self, app):
        self.app = app

        ui.loadUi()

        # INIT FORMS
        self.computationForm = None
        self._detectorForm = None
        self._scatteringForm = None

        # STYLE
        self.app.setStyle('Plastique')
        with open('xicam/gui/style.stylesheet', 'r') as f:
            self.app.setStyleSheet(f.read())


        # SETUP FEATURES
        featuremanager.layout = ui.mainwindow.featuresList
        featuremanager.load()




        # INIT EXPERIMENT
        self.newExperiment()



        # WIREUP CONTROLS
        ui.mainwindow.addFeatureButton.clicked.connect(featuremanager.addLayer)
        ui.mainwindow.addSubstrateButton.clicked.connect(featuremanager.addSubstrate)
        ui.mainwindow.addParticleButton.clicked.connect(featuremanager.addParticle)
        ui.mainwindow.showScatteringButton.clicked.connect(self.showScattering)
        ui.mainwindow.showComp50utationButton.clicked.connect(self.showComputation)
        ui.mainwindow.showDetectorButton.clicked.connect(self.showDetector)
        ui.mainwindow.addParticleButton.setMenu(ui.particlemenu)
        ui.mainwindow.runLocal.clicked.connect(self.runLocal)


        # SETUP DISPLAY
        display.load()
        ui.mainwindow.latticeplaceholder.addWidget(display.viewWidget)


        # END
        ui.mainwindow.show()
        ui.mainwindow.raise_()


    def newExperiment(self):
        pass
        # featuremanager.clearFeatures()
        # featuremanager.addSubstrate()
        # featuremanager.addLayer()

    def showFeature(self, index):
        self.showForm(index.internalPointer().form)


    def showComputation(self):
        if self.computationForm is None:
            self.computationForm = featuremanager.loadform('xicam/gui/guiComputation.ui')
        self.showForm(self.computationForm)

    def showScattering(self):
        self.showForm(self.scatteringForm.form)

    def showDetector(self):
        self.showForm(self.detectorForm.form)

    def showForm(self, form):
        ui.mainwindow.featureWidgetHolder.addWidget(form)
        ui.mainwindow.featureWidgetHolder.setCurrentWidget(form)

    @property
    def detectorForm(self):
        if self._detectorForm is None:
            self._detectorForm = customwidgets.detector()
        return self._detectorForm

    @property
    def scatteringForm(self):
        if self._scatteringForm is None:
            self._scatteringForm = customwidgets.scattering()
        return self._scatteringForm

    # def runLocal(self):
    #     shapes = []
    #     layers = []
    #
    #     shapes = [feature.toDict() for feature in featuremanager.features if type(feature) is customwidgets.particle]
    #     layers = [feature.toDict() for feature in featuremanager.features if
    #               type(feature) is customwidgets.layer or customwidgets.substrate]
    #     unitcells = [feature.structure.toUnitCellDict() for feature in featuremanager.features if
    #                  type(feature) is customwidgets.particle]
    #     structures = [feature.structure.toStructureDict() for feature in featuremanager.features if
    #                   type(feature) is customwidgets.particle]
    #
    #     out = {'hipGisaxsInput': UnsortableOrderedDict([('shapes', shapes),
    #                                                     ('unitcells', unitcells),
    #                                                     ('layers', layers),
    #                                                     ('structures', structures),
    #                                                     ('instrumentation', self.detectorForm.toDict()),
    #                                                     ('computation', self.scatteringForm.toDict())])}
    #     with open('test.json', 'w') as outfile:
    #         json.dump(out, outfile, indent=4)
    #
    #     with open('test.yml', 'w') as outfile:
    #         yaml.dump(out, outfile, indent=4)
    #
    #     print yaml.dump(out, indent=4)







from PySide import QtCore
from PySide.QtCore import Qt
import featuremanager


class featuresModel(QtCore.QAbstractListModel):
    """
    This model displays items for each feature
    """

    def __init__(self):
        QtCore.QAbstractListModel.__init__(self)


    def widgetchanged(self):
        self.modelReset.emit()

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(featuremanager.features)

    def data(self, index, role=Qt.DisplayRole):
        if role == Qt.DisplayRole:
            return featuremanager.features[index.row()].name

            # The view is asking for the actual data, so, just return the item it's asking for.
            # return self._items[index.row()]
            # elif role == Qt.BackgroundRole:
            # Here, it's asking for some background decoration.
            # Let's mix it up a bit: mod the row number to get even or odd, and return different
            # colours depending.
            #if index.row() % 2 == 0:
            #    return QColor(Qt.gray)
            #else:
            #    return QColor(Qt.darkGray)
        else:
            # We don't care about anything else, so make sure to return an empty QVariant.
            return None


class FeatureTree(QtCore.QAbstractItemModel):
    def __init__(self, ):
        QtCore.QAbstractItemModel.__init__(self)
        self._root = featuremanager.experiment()

    def rowCount(self, in_index):
        if in_index.isValid():
            return in_index.internalPointer().childCount()
        return self._root.childCount()

    def addChild(self, in_node, in_parent):
        if not in_parent or not in_parent.isValid():
            parent = self._root
        else:
            parent = in_parent.internalPointer()
        parent.addChild(in_node)

    def index(self, row, column, parent=None):
        if not parent.isValid():
            return self.createIndex(row, column, self._root)
        parentNode = parent.internalPointer()
        return self.createIndex(row, column, parentNode.child(row))

    def parent(self, in_index):
        if in_index.isValid():
            p = in_index.internalPointer().parent()
            if p:
                return QtCore.QAbstractItemModel.createIndex(self, p.row(), 0, p)
        return QtCore.QModelIndex()

    def columnCount(self, in_index):
        if in_index.isValid():
            return in_index.internalPointer().columnCount()
        return self._root.childCount()

    def data(self, in_index, role):
        if not in_index.isValid():
            return None
        node = in_index.internalPointer()
        if role == QtCore.Qt.DisplayRole:
            return node.name
        return None

    def headerData(self, section, orientation, role):
        if orientation == Qt.Horizontal and role == Qt.DisplayRole and section == 0:
            return 'Name'
        return None


from PySide import QtCore


class InvokeEvent(QtCore.QEvent):
    EVENT_TYPE = QtCore.QEvent.Type(QtCore.QEvent.registerEventType())

    def __init__(self, fn, *args, **kwargs):
        QtCore.QEvent.__init__(self, InvokeEvent.EVENT_TYPE)
        self.fn = fn
        self.args = args
        self.kwargs = kwargs


class Invoker(QtCore.QObject):
    def event(self, event):
        event.fn(*event.args, **event.kwargs)
        return True

_invoker = Invoker()


def invoke_in_main_thread(fn, *args, **kwargs):
    # print 'attempt invoke:',fn,args,kwargs
    QtCore.QCoreApplication.postEvent(_invoker,
        InvokeEvent(fn, *args, **kwargs))
    # print 'successful invoke:',fn,args,kwargs




__author__ = "Luis Barroso-Luque, Holden Parks"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque",
               "Holden Parks", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


import numpy as np
import numexpr as ne
import concurrent.futures as cf
import tomopy
from tomopy.util import mproc
import scipy.ndimage.filters as snf
import skimage.transform as st


DTYPE_RANGE = {'uint8': (0, 255),
               'uint16': (0, 65535),
               'int8': (-128, 127),
               'int16': (-32768, 32767),
               'float32': (-1, 1),
               'float64': (-1, 1)}

def crop(arr, p11, p12, p21, p22, axis=0):
    """
    Crops a 3D array along a given axis. Equivalent to slicing the array

    Parameters
    ----------
    arr : ndarray
    p11 : int
        First point along first axis
    p12 : int
        Second point along first axis
    p21 : int
        First point along second axis
    p22 : int
        Second point along second axis
    axis : int
        Axis to crop along

    Returns
    -------
    ndarray:
        Cropped array
    """

    slc = []
    pts = [p11, p12, p21, p22]
    for n in range(len(arr.shape)):
        if n == axis:
            slc.append(slice(None))
        else:
            slc.append(slice(int(pts.pop(0)), -int(pts.pop(0))))
    return arr[slc]


def convert_data(arr, imin=None, imax=None, dtype='uint8', intcast='float32'):
    """
    Convert an image or 3D array to another datatype using numexpr

    :param arr: ndarray, data array
    :param dtype: dataype keyword
    :param imin,
    :param imax,
    :param intcast: datatype to cast ints to
    :return: ndarry, converted to dtype

    Parameters
    ----------
    arr : ndarray, data array
    imin : int/float, optional
        Minimum bound of input array. If not given the minimum value of the array is used
    imax : int/float, optional
        Maximum bound of input array. If not given the maximum value of the array is used
    dtype : str
        Dataype keyword. See DTYPE_RANGE keys
    intcast : str
        intermediate cast type if casting to ints (numexpr will complain otherwise)

    Returns
    -------
    ndarray
        Converted array
    """

    allowed_dtypes = ('uint8', 'uint16', 'int8', 'int16', 'float32', 'float64')
    if dtype not in allowed_dtypes:
        raise ValueError('dtype keyword {0} not in allowed keywords {1}'.format(dtype, allowed_dtypes))

    # Determine range to cast values
    minset=False
    if imin is None:
        imin = np.min(arr)
        minset=True
    maxset=False
    if imax is None:
        imax = np.max(arr)
        maxset=True

    np_cast = getattr(np, str(arr.dtype))
    imin, imax =  np_cast(imin),  np_cast(imax)

    # Determine range of new dtype
    omin, omax = DTYPE_RANGE[dtype]
    omin = 0 if imin >= 0 else omin
    omin, omax = np_cast(omin), np_cast(omax)

    if arr.dtype in [np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16,
                     np.uint32, np.uint64, np.bool_, np.int_, np.intc, np.intp]:
        int_cast = getattr(np, str(intcast))
        out = np.empty(arr.shape, dtype=int_cast)
        imin = int_cast(imin)
        imax = int_cast(imax)
        df = int_cast(imax) - int_cast(imin)
    else:
        out = np.empty(arr.shape, dtype=arr.dtype)
        df = imax - imin
    if not minset:
        if np.min(arr) < imin:
            arr = ne.evaluate('where(arr < imin, imin, arr)', out=out)
    if not maxset:
        if np.max(arr) > imax:
            arr = ne.evaluate('where(arr > imax, imax, arr)', out=out)
    ne.evaluate('(arr - imin) / df', truediv=True, out=out)
    ne.evaluate("out * (omax - omin) + omin", out=out)

    # Cast data to specified type
    return out.astype(np.dtype(dtype), copy=False)


# series of wrappers for simple array operations to expose in workflow pipeline GUI
def array_operation_add(arr, value=0):
    return ne.evaluate('arr + value')

def array_operation_sub(arr, value=0):
    return ne.evaluate('arr - value', truediv=True)

def array_operation_mult(arr, value=1):
    return ne.evaluate('arr * value')

def array_operation_div(arr, value=1):
    return ne.evaluate('arr / value')

def array_operation_max(arr, value=0):
    return np.maximum(arr, value)


def reader(start_sinogram=0, end_sinogram=0, step_sinogram=1, start_projection=0, end_projection=0, step_projection=1,
           sinograms_per_chunk=0, projections_per_chunk=0):

    """
    Function to expose input tomography array parameters in function pipeline GUI

    Parameters
    ----------
    start_sinogram : int
        Start for sinogram processing
    end_sinogram : int
        End for sinogram processing
    step_sinogram : int
        Step for sinogram processing
    start_projection : int
        Start for projection processing
    end_projection : int
        End for projection processing
    step_projection : int
        Step for projection processing
    sinograms_per_chunk : int
        Number of sinograms processed at one time. Limited by machine memory size

    Returns
    -------
    tuple, tuple, int:
        Arguments rearranged into tuples to feed into reconstruction
    """

    return (start_projection, end_projection, step_projection), (start_sinogram, end_sinogram, step_sinogram), \
           sinograms_per_chunk, projections_per_chunk

def slicer(arr, p11=0, p12=0, p21=0, p22=0, p31=0, p32=0):
    """
    Slices a 3D array according the points given

    Parameters
    ----------
    arr : ndarray
    p11 : int
        First point along first axis
    p12 : int
        Second point along first axis
    p21 : int
        First point along second axis
    p22 : int
        Second point along second axis
    p31 : int
        First point along third axis
    p32 : int
        Second point along third axis
    axis : int
        Axis to crop along

    Returns
    -------
    ndarray:
        Cropped array
    """

    slc = []
    pts = [p11, p12, p21, p22, p31, p32]
    for n in range(len(arr.shape)):
        slc.append(slice(pts.pop(0), pts.pop(0)))
    return arr[slc]


def remove_outlier1d(arr, dif, size=3, axis=0, ncore=None, out=None):
    """
    Remove high intensity bright spots (and dark spots) from an array, using a one-dimensional
    median filter along the specified axis

    Parameters
    ----------
    arr : ndarray
        Input array.
    dif : float
        Expected difference value between outlier value and
        the median value of the array.
    size : int
        Size of the median filter.
    axis : int, optional
        Axis along which median filtering is performed.
    ncore : int, optional
        Number of cores that will be assigned to jobs.
    out : ndarray, optional
        Output array for result.  If same as arr, process will be done in-place.
    Returns
    -------
    ndarray
       Corrected array.
    """
    arr = arr.astype(np.float32, copy=False)
    dif = np.float32(dif)

    tmp = np.empty_like(arr)

    other_axes = [i for i in range(arr.ndim) if i != axis]
    largest = np.argmax([arr.shape[i] for i in other_axes])
    lar_axis = other_axes[largest]
    ncore, chnk_slices = mproc.get_ncore_slices(arr.shape[lar_axis], ncore=ncore)
    filt_size = [1] * arr.ndim
    filt_size[axis] = size

    with cf.ThreadPoolExecutor(ncore) as e:
        slc = [slice(None)] * arr.ndim
        for i in range(ncore):
            slc[lar_axis] = chnk_slices[i]
            e.submit(snf.median_filter, arr[slc], size=filt_size, output=tmp[slc], mode='mirror')

    with mproc.set_numexpr_threads(ncore):
        out = ne.evaluate('where(abs(arr-tmp)>=dif,tmp,arr)', out=out)

    return out

def beam_hardening(arr, a0=0, a1=1.0, a2=0, a3=0, a4=0, a5=0.1):
    """
    beam hardening correction, based on "Correction for beam hardening in computed tomography",
    Gabor Herman, 1979 Phys. Med. Biol. 24 81

    Correction is: tomo = a0 + a1*tomo + a2*tomo^2 + a3*tomo^3 + a4*tomo^4 + a5*tomo^5
    """

    loc_dict = {}
    loc_dict['a0'] = np.float32(a0)
    loc_dict['a1'] = np.float32(a1)
    loc_dict['a2'] = np.float32(a2)
    loc_dict['a3'] = np.float32(a3)
    loc_dict['a4'] = np.float32(a4)
    loc_dict['a5'] = np.float32(a5)

    return ne.evaluate('a0 + a1*tomo + a2*tomo**2 + a3*tomo**3 + a4*tomo**4 + a5*tomo**5', local_dict=loc_dict)

def correct_tilt(arr, tilt=0, tiltcenter_slice=None, tiltcenter_det=None, sino_0=0):
    """
    Offset dataset tilt

    Parameters
    ----------
    arr : ndarray
        Input array.
    tilt :
    tiltcenter_slice : int, optional
        Center of dataset in x (sinogram) direction
    tiltcenter_det: int, optional
        Center of dataset in y (image height) direction
    sino_0 : int, optional
        Position of first sinogram in 'arr' relative to larger dataset. For example, if sino_0=200, then the first
        sinogram in 'arr' is the 200th in the larger dataset from which 'arr' is derived
    Returns
    -------
    ndarray
       Corrected array.
    """

    if not tiltcenter_slice:
        tiltcenter_slice = arr.shape[1]/2
    if not tiltcenter_det:
        tiltcenter_det = arr.shape[2]/2

    new_center = tiltcenter_slice - 0.5 -sino_0
    center_det = tiltcenter_det - 0.5
    cntr = (center_det, new_center)
    for b in range(arr.shape[0]):
        arr[b] = st.rotate(arr[b], tilt, center=cntr, preserve_range=True, order=1, mode='edge', clip=True)

    return arr

def sino_360_to_180(data, overlap=0, rotation='left'):
    """
    Wrapper for 360_to_180 function (see below) to handle even/odd shaped data
    """

    if data.shape[0]%2>0:
        return do360_to_180(data[0:-1,:,:], overlap=overlap, rotation=rotation)
    else:
        return do360_to_180(data[:,:,:], overlap=overlap, rotation=rotation)


def do360_to_180(data, overlap=0, rotation='left'):
    """
    Converts 0-360 degrees sinogram to a 0-180 sinogram.

    Parameters
    ----------
    data : ndarray
        Input 3D data.

    overlap : scalar, optional
        Overlapping number of pixels.

    rotation : string, optional
        Left if rotation center is close to the left of the
        field-of-view, right otherwise.

    Returns
    -------
    ndarray
    Output 3D data.
    """
    dx, dy, dz = data.shape
    lo = overlap // 2
    ro = overlap - lo
    n = dx // 2
    out = np.zeros((n, dy, 2*dz - overlap), dtype=data.dtype)
    if rotation == 'left':
        weights = (np.arange(overlap) + 0.5) / overlap
        out[:, :, -dz + overlap:] = data[:n, :, overlap:]
        out[:, :, :dz - overlap] = data[n:2 * n, :, overlap:][:, :, ::-1]
        out[:, :, dz - overlap:dz] = weights * data[:n, :, :overlap] + (weights * data[n:2 * n, :, :overlap])[:, :,
                                                                       ::-1]
    elif rotation == 'right':
        weights = (np.arange(overlap)[::-1] + 0.5) / overlap
        out[:, :, :dz - overlap] = data[:n, :, :-overlap]
        out[:, :, -dz + overlap:] = data[n:2 * n, :, :-overlap][:, :, ::-1]
        out[:, :, dz - overlap:dz] = weights * data[:n, :, -overlap:] + (weights * data[n:2 * n, :, -overlap:])[:, :,
                                                                        ::-1]
    return out

def normalize(tomo, flats, dark, cutoff=None, ncore=None):
    """
    Wrapper for tomopy.normalize.normalize function (make more similar to nearest flats norm function)
    """

    return tomopy.normalize(arr=tomo, flat=flats, dark=dark, cutoff=cutoff, ncore=ncore)


if __name__ == '__main__':
    import tomopy
    from matplotlib.pyplot import imshow, show, figure
    d = tomopy.read_als_832h5('/home/lbluque/TestDatasetsLocal/dleucopodia.h5', ind_tomo=(500, 501, 502, 503))
    # d = np.array(d[0], dtype=np.float32)
    d = d[0]
    c = convert_data(d, imin=0, imax=0)
    print(d.dtype)
    print(c.dtype)
    figure(0)
    imshow(d[0])
    figure(1)
    imshow(c[0])
    show()


# --coding: utf-8 --
import pyqtgraph.parametertree.parameterTypes as pTypes
from pyqtgraph.parametertree import Parameter, ParameterTree, ParameterItem, registerParameterType
from PySide import QtCore, QtGui
import pyqtgraph as pg
import spacegrp_peaks
import numpy as np
from xicam import config
import msg

# TODO: Add index of refraction to interface and backend
# TODO: Add q,twotheta, alphaf to tooltip
# TODO: check scaling
# TODO: Actively determine number of orders to display
# TODO: Connect orientation interface with backend
# TODO: Align overlay with center, mirror:


try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    def _fromUtf8(s):
        return s


class vector(QtGui.QWidget):
    sigValueChanged = QtCore.Signal()
    sigChanged = sigValueChanged


    def __init__(self):
        super(vector, self).__init__()

        self.horizontalLayout = QtGui.QHBoxLayout(self)
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout.setObjectName(_fromUtf8("horizontalLayout"))
        self.UnitCellVec1LeftParenthesis3D = QtGui.QLabel(self)
        font = QtGui.QFont()
        font.setPointSize(30)
        self.UnitCellVec1LeftParenthesis3D.setFont(font)
        self.UnitCellVec1LeftParenthesis3D.setObjectName(_fromUtf8("UnitCellVec1LeftParenthesis3D"))
        self.horizontalLayout.addWidget(self.UnitCellVec1LeftParenthesis3D)
        self.value1 = QtGui.QDoubleSpinBox(self)
        self.value1.setDecimals(1)
        self.value1.setMinimum(-1000.0)
        self.value1.setMaximum(1000.0)
        self.value1.setSingleStep(0.5)
        self.value1.setProperty("value", 0.0)
        self.value1.setObjectName(_fromUtf8("value1"))
        self.horizontalLayout.addWidget(self.value1)
        self.UnitCellVec1Comma1 = QtGui.QLabel(self)
        self.UnitCellVec1Comma1.setObjectName(_fromUtf8("UnitCellVec1Comma1"))
        self.horizontalLayout.addWidget(self.UnitCellVec1Comma1)
        self.value2 = QtGui.QDoubleSpinBox(self)
        self.value2.setDecimals(1)
        self.value2.setMinimum(-1000.0)
        self.value2.setMaximum(1000.0)
        self.value2.setSingleStep(0.5)
        self.value2.setObjectName(_fromUtf8("value2"))
        self.horizontalLayout.addWidget(self.value2)
        self.UnitCellVec1Comma2 = QtGui.QLabel(self)
        self.UnitCellVec1Comma2.setObjectName(_fromUtf8("UnitCellVec1Comma2"))
        self.horizontalLayout.addWidget(self.UnitCellVec1Comma2)
        self.value3 = QtGui.QDoubleSpinBox(self)
        self.value3.setDecimals(1)
        self.value3.setMinimum(-1000.0)
        self.value3.setMaximum(1000.0)
        self.value3.setSingleStep(0.5)
        self.value3.setObjectName(_fromUtf8("value3"))
        self.horizontalLayout.addWidget(self.value3)
        self.UnitCellVec1RightParenthesis3D = QtGui.QLabel(self)
        self.UnitCellVec1RightParenthesis3D.setFont(font)
        self.UnitCellVec1RightParenthesis3D.setObjectName(_fromUtf8("UnitCellVec1RightParenthesis3D"))
        self.horizontalLayout.addWidget(self.UnitCellVec1RightParenthesis3D)

        self.UnitCellVec1LeftParenthesis3D.setText("(")
        self.UnitCellVec1Comma1.setText(",")
        self.UnitCellVec1Comma2.setText(",")
        self.UnitCellVec1RightParenthesis3D.setText(")")

        self.value1.valueChanged.connect(self.sigValueChanged)
        self.value2.valueChanged.connect(self.sigValueChanged)
        self.value3.valueChanged.connect(self.sigValueChanged)

    def value(self):
        return self.value1.value(), self.value2.value(), self.value3.value()


    def setValue(self, v):
        self.value1.setValue(v[0])
        self.value2.setValue(v[1])
        self.value3.setValue(v[2])


    def setEnabled(self, enabled):
        self.value1.setEnabled(enabled)
        self.value2.setEnabled(enabled)
        self.value3.setEnabled(enabled)


class VectorParameterItem(pTypes.WidgetParameterItem):
    def makeWidget(self):
        w = vector()
        opts = self.param.opts
        value = opts.get('value', None)
        if value is not None:
            w.setValue(value)

        self.value = w.value
        self.setValue = w.setValue

        return w

    def valueChanged(self, *args, **kwargs):
        super(VectorParameterItem, self).valueChanged(*args, **kwargs)


class VectorParameter(Parameter):
    itemClass = VectorParameterItem

    def __init__(self, *args, **kwargs):
        super(VectorParameter, self).__init__(*args, **kwargs)


    def defaultValue(self):
        return (0, 0, 0)


# class peak(object):
#     def __init__(self, mode, hkl, x, y, twotheta=None, alphaf=None, q=None):
#         self.mode = mode # either 'Transmission' or 'Reflection'
#         self.hkl = hkl
#         self.x = x
#         self.y = y
#         self.twotheta = twotheta
#         self.alphaf = alphaf
#         self.q = q
#
#
#
#     def __str__(self):
#         s = u"Peak type: {}\n".format(self.mode)
#         s += u"Lattice vector (h,k,l): {}\n".format(self.hkl)
#         if self.twotheta is not None: s += u"2\u03B8: {}\n".format(self.twotheta)
#         if self.alphaf is not None: s += u"\u03B1f: {}\n".format(self.alphaf)
#         if self.q is not None: s += u"q: {}".format(self.q)
#         return s


class peakoverlay(pg.ScatterPlotItem):
    def __init__(self, peaks):
        self.peaks = peaks
        self.centerx = 0
        self.centery = 0
        if len(peaks):
            x, y = zip(*[[p.x, p.y] for p in peaks])
            symbols = ['s' if p.mode == 'Transmission' else 'o' for p in peaks]
            colors = [pg.mkPen(0, 255, 0, 255) if p.mode == 'Transmission' else pg.mkPen(255, 0, 255, 255) for p in peaks]
            super(peakoverlay, self).__init__(x, y, size=10, brush=None, pen=colors, symbol=symbols)
            self.display_text = pg.TextItem(text='', color=(255, 255, 255), anchor=(0, 1),
                                            fill=pg.mkBrush(255, 127, 0, 100))
            self.display_text.hide()

            # if parent is not None:
            # parent.addItem(self.scatterPoints)
            #     parent.addItem(self.display_text)

    def setCenter(self,x,y):
        self.centerx=x
        self.centery=y
        self.displayRelative()

    def displayRelative(self):
        px, py = zip(*[[p.x, p.y] for p in self.peaks])

        symbols = ['s' if p.mode == 'Transmission' else 'o' for p in self.peaks]
        colors = [pg.mkPen(0, 255, 0, 255) if p.mode == 'Transmission' else pg.mkPen(255, 0, 255, 255) for p in self.peaks]
        self.setData(x=np.array(px)+self.centerx,y=np.array(py)+self.centery,size=10,brush=None,pen=colors,symbol=symbols)

    def enable(self, parent):
        self.scene().sigMouseMoved.connect(self.onMove)
        parent.addItem(self.display_text)


    # def onMove(self,pos):
    # act_pos = self.scatterPoints.mapFromScene(pos)
    #     p1 = self.scatterPoints.pointsAt(act_pos)
    #     if len(p1)!=0:
    #         self.display_text.setText(str(self))
    #         self.display_text.setPos(self.x,self.y)
    #         self.display_text.show()
    #     else:
    #         self.display_text.hide()

    def peaksAtRelative(self, pos):
        x = pos.x()-self.centerx
        y = pos.y()-self.centery
        pw = self.pixelWidth()
        ph = self.pixelHeight()
        peaks = []
        ss=self.points()[0].size()
        for p in self.peaks:
            sx,sy = p.pos()
            s2x = s2y = ss * 0.5
            if self.opts['pxMode']:
                s2x *= pw
                s2y *= ph
            if x > sx - s2x and x < sx + s2x and y > sy - s2y and y < sy + s2y:
                peaks.append(p)
                # print "HIT:", x, y, sx, sy, s2x, s2y
                # else:
                # print "No hit:", (x, y), (sx, sy)
                # print "       ", (sx-s2x, sy-s2y), (sx+s2x, sy+s2y)
        # pts.sort(lambda a,b: cmp(b.zValue(), a.zValue()))
        return peaks[::-1]

    def onMove(self, pos):
        pos = self.mapFromScene(pos)
        peaks = self.peaksAtRelative(pos)



        if len(peaks) != 0:
            s = u''
            #print points
            for peak in peaks:
                    if s != u'': s += '\n\n'
                    s += unicode(peak)

            self.display_text.setText(s)
            self.display_text.setPos(pos)
            self.display_text.show()
            msg.logMessage(s)
        else:
            self.display_text.hide()
            #print str(self)



        if len(peaks) > 6:
            self.display_text.setText('Too many points under cursor.\n'
                                      'Zoom in or check Log info.')
            self.display_text.setPos(pos)
            self.display_text.show()

class spacegroupwidget(ParameterTree):
    sigDrawSGOverlay = QtCore.Signal(peakoverlay)

    def __init__(self):
        super(spacegroupwidget, self).__init__()
        self.parameter = pTypes.GroupParameter(name='')
        self.setParameters(self.parameter, showTop=False)
        self.crystalsystem = pTypes.ListParameter(type='list', name='Crystal System', values=spacegrouptypes,
                                                  value=None)
        self.spacegroupparameter = pTypes.ListParameter(type='list', name='Space Group', values=spacegroupnames[0],
                                                        value=None)
        self.rotationstyle = pTypes.ListParameter(type='list', name='Rotation Mode',
                                                  values=['Sample-frame vector', 'Crystal-frame vector',
                                                          'Crystal plane'])
        self.rotationstyle.sigValueChanged.connect(self.rotationmodechanged)

        self.rotationxyz = VectorParameter(name='Vector (x,y,z)')
        self.rotationvectorsample = hideableGroup(name='Rotation (sample-frame vector)', children=[self.rotationxyz])
        self.rotationxyz.setValue([0,0,1])

        self.rotationuvw = VectorParameter(name='Vector (u,v,w)')
        self.rotationvectorcrystal = hideableGroup(name='Rotation (crystal-frame vector)', children=[self.rotationuvw])
        self.rotationuvw.setValue([0,0,1])

        self.rotationhkl = VectorParameter(name='Vector (h,k,l)')
        self.rotationplane = hideableGroup(name='Rotation (crystal plane)', children=[self.rotationhkl])
        self.rotationhkl.setValue([0,0,1])

        self.spacegroupeditors = [triclinicparameter(), monoclinicparameter(), orthorhombicparameter(),
                                  tetragonalparameter(), trigonalparameter(), hexagonalparameter(), cubicparameter()]
        self.rotations = [self.rotationvectorsample, self.rotationvectorcrystal, self.rotationplane]

        self.delta = pTypes.SimpleParameter(name=u'δ', type='float', value=2.67150153E-06)
        self.beta = pTypes.SimpleParameter(name=u'β', type='float', value=3.71373554E-09)
        self.refractiveindex = pTypes.GroupParameter(name='Refractive Index', children=[self.delta, self.beta])

        self.redrawsg = pTypes.ActionParameter(name='Overlay space group')
        self.redrawsg.sigActivated.connect(self.drawoverlay)

        self.parameter.addChildren([self.crystalsystem, self.spacegroupparameter] + self.spacegroupeditors + [
            self.rotationstyle] + self.rotations + [self.refractiveindex, self.redrawsg])

        self.hidechildren()
        self.hiderotations()
        self.parameter.children()[2].show()
        self.rotations[0].show()
        self.crystalsystem.sigValueChanged.connect(self.crystalsystemchanged)

    def activelatticetype(self):
        return self.spacegroupeditors[self.crystalsystem.reverse[0].index(self.crystalsystem.value())]

    def _getRotationVector(self):
        return [self.rotationxyz,self.rotationuvw,self.rotationhkl][self._getRotationType()].value()

    def _getRotationType(self):
        return self.rotationstyle.reverse[0].index(self.rotationstyle.value())

    def drawoverlay(self):


        activelatticetype = self.activelatticetype()
        SG = self.spacegroupparameter.value()
        refbeta = self.beta.value()
        refdelta = self.delta.value()
        peaks = spacegrp_peaks.find_peaks(float(activelatticetype.a.value()), float(activelatticetype.b.value()),
                                          float(activelatticetype.c.value()), activelatticetype.alpha.value(),
                                          activelatticetype.beta.value(), activelatticetype.gamma.value(),
                                          normal=self._getRotationVector(), norm_type=['xyz','hkl','uvw'][self._getRotationType()], refdelta=refdelta,refbeta=refbeta,order=5,unitcell=None,space_grp=SG)
        for peak in peaks:
            msg.logMessage(unicode(peak),msg.DEBUG)


        #     print key + " -> " + str(peaks[key])
            center = config.activeExperiment.center
            sdd = config.activeExperiment.getvalue('Detector Distance')
            pixelsize = config.activeExperiment.getvalue('Pixel Size X')
            peak.position(center,sdd,pixelsize)
        #     pixels = spacegrp_peaks.angles_to_pixels(np.array(peaks[key]), center, sdd)
        #     peaks[key] = pixels

        # peaks = [peak('Transmission', p, peaks[p][0][0], peaks[p][0][1], 1, 1, 1) for p in peaks if
        #          not np.any(peaks[p] < -100000)] + \
        #         [peak('Reflection', p, peaks[p][1][0], peaks[p][1][1], 1, 1, 1) for p in peaks if
        #          not np.any(peaks[p] < -100000)]

        self.sigDrawSGOverlay.emit(peakoverlay(peaks))
        # self.sigDrawSGOverlay.emit(peakoverlay(
        #    [peak('Reflection', 1, 1, 1, 100, 100, 1, 1, 1), peak('Transmission', 1, 1, 1, 200, 200, 1, 1, 1)]))

    def hidechildren(self):
        for child in self.spacegroupeditors:
            child.hide()

    def hiderotations(self):
        for child in self.rotations:
            child.hide()

    def rotationmodechanged(self, _, value):
        self.hiderotations()
        self.rotations[self.rotationstyle.reverse[0].index(value)].show()

    def crystalsystemchanged(self, _, value):
        self.hidechildren()
        self.parameter.param(value).show()
        self.spacegroupparameter.setLimits(spacegroupnames[spacegrouptypes.index(value)])


class hideableGroupParameterItem(pTypes.GroupParameterItem):
    def optsChanged(self, param, opts):
        super(hideableGroupParameterItem, self).optsChanged(param, opts)
        if 'visible' in opts:
            self.setHidden(not opts['visible'])


class hideableGroup(pTypes.GroupParameter):
    itemClass = hideableGroupParameterItem


class spacegroup(hideableGroup):
    crystalsystem = None

    def __init__(self, name):
        super(spacegroup, self).__init__(name=name)

        self.alpha = pTypes.SimpleParameter(type='float', name=u'α', value=90, step=.01)
        self.beta = pTypes.SimpleParameter(type='float', name=u'β', value=90, step=.01)
        self.gamma = pTypes.SimpleParameter(type='float', name=u'γ', value=90, step=.01)
        self.a = pTypes.SimpleParameter(type='float', name='a', value=10.e-9, step=1.e-9, suffix='m', siPrefix=True)
        self.b = pTypes.SimpleParameter(type='float', name='b', value=10.e-9, step=1.e-9, suffix='m', siPrefix=True)
        self.c = pTypes.SimpleParameter(type='float', name='c', value=10.e-9, step=1.e-9, suffix='m', siPrefix=True)
        self.addChildren([self.alpha, self.beta, self.gamma, self.a, self.b, self.c])

    def _setb(self, _, value):
        self.b.setValue(value, blockSignal=self.b.sigValueChanged)

    def _setc(self, _, value):
        self.c.setValue(value, blockSignal=self.c.sigValueChanged)


class triclinicparameter(spacegroup):
    crystalsystem = 'Triclinic'

    def __init__(self):
        super(triclinicparameter, self).__init__(name=self.crystalsystem)


class monoclinicparameter(spacegroup):
    crystalsystem = 'Monoclinic'

    def __init__(self):
        super(monoclinicparameter, self).__init__(name=self.crystalsystem)
        self.alpha.setValue(90)
        self.gamma.setValue(90)
        self.alpha.setReadonly(True)
        self.gamma.setReadonly(True)


class orthorhombicparameter(spacegroup):
    crystalsystem = 'Orthorhombic'

    def __init__(self):
        super(orthorhombicparameter, self).__init__(name=self.crystalsystem)
        self.alpha.setValue(90)
        self.beta.setValue(90)
        self.gamma.setValue(90)
        self.alpha.setReadonly(True)
        self.beta.setReadonly(True)
        self.gamma.setReadonly(True)


class tetragonalparameter(spacegroup):
    crystalsystem = 'Tetragonal'

    def __init__(self):
        super(tetragonalparameter, self).__init__(name=self.crystalsystem)
        self.alpha.setValue(90)
        self.beta.setValue(90)
        self.gamma.setValue(90)
        self.alpha.setReadonly(True)
        self.beta.setReadonly(True)
        self.gamma.setReadonly(True)
        self.b.setReadonly(True)
        self.a.sigValueChanged.connect(self._setb)


class trigonalparameter(spacegroup):
    crystalsystem = 'Trigonal'

    def __init__(self):
        super(trigonalparameter, self).__init__(name=self.crystalsystem)
        self.alpha.setValue(90)
        self.beta.setValue(90)
        self.gamma.setValue(120)
        self.alpha.setReadonly(True)
        self.beta.setReadonly(True)
        self.gamma.setReadonly(True)
        self.b.setReadonly(True)
        self.a.sigValueChanged.connect(self._setb)


class hexagonalparameter(spacegroup):
    crystalsystem = 'Hexagonal'

    def __init__(self):
        super(hexagonalparameter, self).__init__(name=self.crystalsystem)
        self.alpha.setValue(90)
        self.beta.setValue(90)
        self.gamma.setValue(120)
        self.gamma.setReadonly(True)
        self.b.setReadonly(True)
        self.a.sigValueChanged.connect(self._setb)


class cubicparameter(spacegroup):
    crystalsystem = 'Cubic'

    def __init__(self):
        super(cubicparameter, self).__init__(name=self.crystalsystem)
        self.alpha.setValue(90)
        self.beta.setValue(90)
        self.gamma.setValue(90)
        self.alpha.setReadonly(True)
        self.beta.setReadonly(True)
        self.gamma.setReadonly(True)
        self.b.setReadonly(True)
        self.c.setReadonly(True)
        self.a.sigValueChanged.connect(self._setb)
        self.a.sigValueChanged.connect(self._setc)


triclinicspacegroupnames = ['P1', 'P-1']
monoclinicspacegroupnames = ['P2', u'P2₁', 'Pm', 'Pc', 'Cm', 'Cc', 'P2/m', u'P2₁/m', 'C2/m', 'P2/c', u'P2₁/c', 'C2/c']
orhorhombicspacegroupnames = ['P222', u'P222₁', u'P2₁2₁2', u'P2₁2₁2₁', u'C222₁', 'C222', 'F222', 'I222', u'I2₁2₁2₁',
                              'Pmm2', u'Pmc2₁',
                              'Pcc2', 'Pma2', u'Pca2₁', 'Pnc2', u'Pmn2₁', 'Pba2', u'Pna2₁', 'Pnn2', 'Cmm2', u'Cmc2₁',
                              'Ccc2', 'Amm2',
                              'Aem2', 'Ama2', 'Aea2', 'Fmm2', 'Fdd2', 'Imm2', 'Iba2', 'Ima2', 'Pmmm', 'Pnnn', 'Pccm',
                              'Pban', 'Pmma',
                              'Pnna', 'Pmna', 'Pcca', 'Pbam', 'Pccn', 'Pbcm', 'Pnnm', 'Pmmn', 'Pbcn', 'Pbca', 'Pnma',
                              'Cmcm', 'Cmce',
                              'Cmmm', 'Cccm', 'Cmme', 'Ccce', 'Fmmm', 'Fddd', 'Immm', 'Ibam', 'Ibca', 'Imma']
tetragonalspacegroupnames = ['P4', u'P4₁', u'P4₂', u'P4₃', 'I4', u'I4₁', 'P-4', 'I-4', 'P4/m', u'P4₂/m', 'P4/n',
                             u'P4₂/n', 'I4/m', u'I4₁/a',
                             'P422', u'P42₁2', u'P4₁22', u'P4₁2₁2', u'P4₂22', u'P4₂2₁2', u'P4₃22', u'P4₃2₁2', 'I422',
                             u'I4₁22', 'P4mm',
                             'P4bm', u'P4₂cm', u'P4₂nm', 'P4cc', 'P4nc', 'P4₂mc', 'P4₂bc', 'I4mm', 'I4cm', u'I4₁md',
                             u'I4₁cd', 'P-42m',
                             'P-42c', 'P-42₁m', 'P-42₁c', 'P-4m2', 'P-4c2', 'P-4b2', 'P-4n2', 'I-4m2', 'I-4c2', 'I-42m',
                             'I-42d',
                             'P4/mmm', 'P4/mcc', 'P4/nbm', 'P4/nnc', 'P4/mbm', 'P4/mnc', 'P4/nmm', 'P4/ncc', 'P4₂/mmc',
                             'P4₂/mcm',
                             u'P4₂/nbc', u'P4₂/nnm', u'P4₂/mbc', u'P4₂/mnm', u'P4₂/nmc', u'P4₂/ncm', 'I4/mmm', 'I4/mcm',
                             u'I4₁/amd',
                             u'I4₁/acd']
trigonalspacegroupnames = ['P3', u'P3₁', u'P3₂', 'R3', 'P-3', 'R-3', 'P312', 'P321', u'P3₁12', u'P3₁21', u'P3₂12',
                           u'P3₂21', 'R32', 'P3m1',
                           'P31m', 'P3c1', 'P31c', 'R3m', 'R3c', 'P-31m', 'P-31c', 'P-3m1', 'P-3c1', 'R-3m', 'R-3c']
hexagonalspacegroupnames = ['P6', u'P6₁', u'P6₅', u'P6₂', u'P6₄', u'P6₃', 'P-6', 'P6/m', 'P6₃/m', 'P622', u'P6₁22',
                            u'P6₅22', u'P6₂22',
                            u'P6₄22', u'P6₃22', 'P6mm', 'P6cc', 'P6₃cm', u'P6₃mc', 'P-6m2', 'P-6c2', 'P-62m', 'P-62c',
                            'P6/mmm',
                            'P6/mcc', u'P6₃/mcm', u'P6₃/mmc']
cubicspacegroupnames = ['P23', 'F23', 'I23', u'P2₁3', u'I2₁3', 'Pm-3', 'Pn-3', 'Fm-3', 'Fd-3', 'Im-3', 'Pa-3', 'Ia-3',
                        'P432', u'P4₂32',
                        'F432', u'F4₁32', 'I432', u'P4₃32', u'P4₁32', u'I4₁32', 'P-43m', 'F-43m', 'I-43m', 'P-43n',
                        'F-43c', 'I-43d',
                        'Pm-3m', 'Pn-3n', 'Pm-3n', 'Pn-3m', 'Fm-3m', 'Fm-3c', 'Fd-3m', 'Fd-3c', 'Im-3m', 'Ia-3d']
import sgexclusions
spacegroupnames = [sgexclusions.Triclinic.conditions.keys(), sgexclusions.Monoclinic.conditions.keys(), sgexclusions.Orthorhombic.conditions.keys(),
                   sgexclusions.Tetragonal.conditions.keys(), sgexclusions.Trigonal.conditions.keys(), sgexclusions.Hexagonal.conditions.keys(), sgexclusions.Cubic.conditions.keys()]
spacegrouptypes = ['Triclinic', 'Monoclinic', 'Orthorhombic', 'Tetragonal', 'Trigonal', 'Hexagonal', 'Cubic']



from pipeline import center_approx
import integration
import numpy as np
import peakfinding
from xicam import config

from pyFAI import calibrant
from functools import wraps
import msg
from xicam import threads
from scipy import signal


def calibrationAlgorithm(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        msg.showMessage('Calibrating...')
        msg.showBusy()

        runnable = threads.RunnableIterator(f, iterator_args=args, iterator_kwargs=kwargs,
                                            callback_slot=showProgress, finished_slot=calibrationFinish)
        threads.add_to_queue(runnable)

    return wrapped


def calibrationFinish():
    msg.showMessage('Calibration complete!', 4)
    msg.hideBusy()


def showProgress(value):
    msg.showProgress(value)


@calibrationAlgorithm
def fourierAutocorrelation(dimg, calibrantkey):
    yield 0
    if dimg.transformdata is None:
        return
    yield 20

    config.activeExperiment.center = center_approx.center_approx(dimg.transformdata,dimg.mask)

    yield 40

    radialprofile = integration.pixel_2Dintegrate(dimg, mask=dimg.mask)

    yield 60

    peaks = np.array(peakfinding.findpeaks(np.arange(len(radialprofile)), radialprofile)).T

    yield 80

    peaks = peaks[peaks[:, 1].argsort()[::-1]]

    for peak in peaks:
        if peak[0] > 15 and not np.isinf(peak[1]):  ####This thresholds the minimum sdd which is acceptable
            bestpeak = peak[0]
            # print peak
            break

    yield 85
    N = 1
    # guess which order is the first
    stds=[np.std((peaks[:,0]/(np.arange(len(peaks))+i))[:4]) for i in range(1,5)]
    if min(stds)<10: # threshold accepting peak as higher N
        N=np.argmin(stds)+1 # index of the first detected peak

    msg.logMessage(('Using peak #',N,' for calibration...'))
    yield 90

    calibrant1stpeak = calibrant.ALL_CALIBRANTS[calibrantkey].dSpacing[N-1]

    # Calculate sample to detector distance for lowest q peak
    tth = 2 * np.arcsin(0.5 * config.activeExperiment.getvalue('Wavelength') / calibrant1stpeak / 1.e-10)
    tantth = np.tan(tth)
    sdd = bestpeak * config.activeExperiment.getvalue('Pixel Size X') / tantth

    config.activeExperiment.setvalue('Detector Distance', sdd)

    #        self.refinecenter()

    dimg.invalidatecache()

    yield 100


@calibrationAlgorithm
def rickerWavelets(dimg, calibrantkey):
    yield 0

    if dimg.transformdata is None:
        return

    yield 1

    radii = np.arange(30, 100)
    img = dimg.transformdata

    maxval = 0
    center = np.array([0, 0], dtype=np.int)
    for i in range(len(radii)):
        yield int(96. * i / len(radii)) + 1
        w = center_approx.tophat2(radii[i], scale=1000)
        im2 = signal.fftconvolve(img, w, 'same')
        if im2.max() > maxval:
            maxval = im2.max()
            center = np.array(np.unravel_index(im2.argmax(), img.shape))

    config.activeExperiment.center = center

    yield 97

    radialprofile = integration.pixel_2Dintegrate(dimg, mask=dimg.mask)

    yield 98

    peaks = np.array(peakfinding.findpeaks(np.arange(len(radialprofile)), radialprofile)).T

    yield 99

    peaks = peaks[peaks[:, 1].argsort()[::-1]]

    for peak in peaks:
        if peak[0] > 25 and not np.isinf(peak[1]):  ####This thresholds the minimum sdd which is acceptable
            bestpeak = peak[0]
            # print peak
            break

    calibrant1stpeak = calibrant.ALL_CALIBRANTS[calibrantkey].dSpacing[0]

    # Calculate sample to detector distance for lowest q peak
    tth = 2 * np.arcsin(0.5 * config.activeExperiment.getvalue('Wavelength') / calibrant1stpeak / 1.e-10)
    tantth = np.tan(tth)
    sdd = bestpeak * config.activeExperiment.getvalue('Pixel Size X') / tantth

    config.activeExperiment.setvalue('Detector Distance', sdd)

    #        self.refinecenter()

    dimg.invalidatecache()

    yield 100


import saxs_calibration as sc
@calibrationAlgorithm
def dpdakRefine(dimg,calibrantkey):

    if dimg.transformdata is None:
        return

    # Refine calibration
    # d-spacing for Silver Behenate
    d_spacings = np.array(sorted(calibrant.ALL_CALIBRANTS[calibrantkey].dSpacing,key=float,reverse=True))

    geometry = config.activeExperiment.getAI()

    data = dimg.rawdata

    print 'Start parameter:'
    print geometry.getFit2D()

    fit_param = ['distance', 'rotation', 'tilt', 'center_x', 'center_y']
    # calculate maxima for every d_spacing
    center = (geometry.getFit2D()['centerX'],
              geometry.getFit2D()['centerY'])
    radial_pos = sc.radial_array(center, data.shape)
    x_data, y_data, = [], []

    # calculate maxima for every d_spacing
    for i in range(len(d_spacings)):
        yield 100/len(d_spacings)*i
        maxima_x, maxima_y, radial_pos = sc.ring_maxima(geometry,
                                                     d_spacings[i],
                                                     data,
                                                     radial_pos,
                                                     10) # TODO: parameterize this
        x_data.extend(maxima_x)
        y_data.extend(maxima_y)



    # start fit
    sc.fit_geometry(geometry,
                        (np.array(x_data), np.array(y_data)),
                        d_spacings,
                        fit_param)


    print 'Final parameter:'
    print geometry.getFit2D()

    config.activeExperiment.setvalue('Detector Distance', geometry.get_dist())
    config.activeExperiment.center = geometry.getFit2D()['centerX'],geometry.getFit2D()['centerY']
    yield 100




__author__ = "Luis Barroso-Luque, Holden Parks"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque",
               "Holden Parks", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"


import os
import numpy as np
import inspect
from collections import OrderedDict
import yaml
from pipeline import msg
from modpkgs import yamlmod

PARAM_TYPES = {'int': int, 'float': float}

# Load yaml with names of all available functions in pipeline
with open('xicam/yaml/tomography/functions.yml','r') as stream:
    funcs=yaml.load(stream)

# load various function dictionaries from function_info.yml file
parameters = {}; als832defaults = {}; aps_defaults = {}; names = {}; function_defaults = {}
with open('xicam/yaml/tomography/functions_info.yml', 'r') as stream:
    info = yaml.load(stream)
    for key in info.keys():
        # load parameter data for available functions
        if 'parameters' in info[key].keys():
            parameters[key] = info[key]['parameters']

        # load dictionary with function parameters to be retrieved from metadata
        try:
            als832defaults[key] = info[key]['conversions']['als']
        except KeyError:
            pass
        try:
            aps_defaults[key] = info[key]['conversions']['aps']
        except KeyError:
            pass

        # load dictionary with pipeline names and function names
        if 'name' in info[key].keys():
            names[key] = info[key]['name']

        # load dictionary of set defaults
        if 'defaults' in info[key].keys():
            function_defaults[key] = info[key]['defaults']

# Add reconstruction methods to function name dictionary, but include the package the method is in
for algorithm in funcs['Functions']['Reconstruction']['TomoPy']:
    names[algorithm] = ['recon', 'tomopy']

for algorithm in funcs['Functions']['Reconstruction']['Astra']:
    names[algorithm] = ['recon', 'astra']

for algorithm in funcs['Functions']['Reconstruction']['TomoCam']:
    names[algorithm] = ['recon','mbir']


def load_pipeline(yaml_file):
    """
    Load a workflow pipeline from a yaml file
    """

    with open(yaml_file, 'r') as y:
        pipeline = yamlmod.ordered_load(y)
    return pipeline


def save_function_pipeline(pipeline, file_name):
    """
    Save a workflow pipeline from dict

    Parameters
    ----------
    pipeline : dict
        dictionary specifying the workflow pipeline
    file_name : str
        file name to save as yml
    """

    if file_name != '':
        file_name = file_name.split('.')[0] + '.yml'
        with open(file_name, 'w') as y:
            yamlmod.ordered_dump(pipeline, y)


def set_als832_defaults(mdata, funcwidget_list, path, shape):
    """
    Set defaults for ALS Beamline 8.3.2 from dataset metadata

    Parameters
    ----------
    mdata : dict
        dataset metadata
    funcwidget_list : list of FunctionWidgets
        list of FunctionWidgets exposed in the UI workflow pipeline
    path: str
        path to dataset
    shape: tuple
        tuple containing dataset shape
    """
    from psutil import cpu_count
    for f in funcwidget_list:
        if f is None:
            continue
        if f.subfunc_name in als832defaults:
            for p in f.params.children():
                if p.name() in als832defaults[f.subfunc_name]:
                    try:
                        v = mdata[als832defaults[f.subfunc_name][p.name()]['name']]
                        t = PARAM_TYPES[als832defaults[f.subfunc_name][p.name()]['type']]
                        v = t(v) if t is not int else t(float(v))  # String literals for ints should not have 0's
                        if 'conversion' in als832defaults[f.subfunc_name][p.name()]:
                            v *= als832defaults[f.subfunc_name][p.name()]['conversion']
                        p.setDefault(v)
                        p.setValue(v)
                    except KeyError as e:
                        msg.logMessage('Key {} not found in metadata. Error: {}'.format(p.name(), e.message),
                                       level=40)
        elif f.func_name == 'Reader': #dataset specific read values
            set_reader_defaults(f, shape, cpu_count())

        elif f.func_name == 'Write': #dataset specific write values
            data_folders = {'bl832data-raw':'bl832data-scratch', 'data-raw':'data-scratch'}
            file_name = path.split("/")[-1].split(".")[0]
            working_dir = path.split(file_name)[0]
            for key in data_folders.keys():
                if key in working_dir:
                    user = working_dir.split('/' + key)[-1].split('/')[1]
                    mount = working_dir.split(key)[0]
                    working_dir = os.path.join(mount, data_folders[key], user)
            outname = os.path.join(working_dir, *2*('RECON_' + file_name,))
            f.params.child('parent folder').setValue(working_dir)
            f.params.child('parent folder').setDefault(working_dir)
            f.params.child('folder name').setValue('RECON_' + file_name)
            f.params.child('folder name').setDefault('RECON_' + file_name)
            f.params.child('file name').setValue('RECON_' + file_name)
            f.params.child('file name').setDefault('RECON_' + file_name)
            f.params.child('fname').setValue(outname)
            f.params.child('fname').setDefault(outname)
        if f.input_functions:
            set_als832_defaults(mdata, funcwidget_list=f.input_functions.values(), path=path, shape=shape)

def set_aps_defaults(mdata, funcwidget_list, path, shape):
    """
    Set defaults for ALS Beamline 8.3.2 from dataset metadata

    Parameters
    ----------
    mdata : dict
        dataset metadata
    funcwidget_list : list of FunctionWidgets
        list of FunctionWidgets exposed in the UI workflow pipeline
    path: str
        path to dataset
    shape: tuple
        tuple containing dataset shape
    """
    from psutil import cpu_count
    for f in funcwidget_list:
        if f is None:
            continue
        if aps_defaults and f.subfunc_name in aps_defaults:
            for p in f.params.children():
                if p.name() in aps_defaults[f.subfunc_name]:
                    try:
                        v = mdata[aps_defaults[f.subfunc_name][p.name()]['name']]
                        t = PARAM_TYPES[aps_defaults[f.subfunc_name][p.name()]['type']]
                        v = t(v) if t is not int else t(
                            float(v))  # String literals for ints should not have 0's
                        if 'conversion' in aps_defaults[f.subfunc_name][p.name()]:
                            v *= aps_defaults[f.subfunc_name][p.name()]['conversion']
                        p.setDefault(v)
                        p.setValue(v)
                    except KeyError as e:
                        msg.logMessage('Key {} not found in metadata. Error: {}'.format(p.name(), e.message),
                                       level=40)
        elif f.func_name == 'Reader':  # dataset specific read values
            set_reader_defaults(f, shape, cpu_count())

        elif f.func_name == 'Padding':
            pad = int(np.ceil((shape[2] * np.sqrt(2) - shape[2]) / 2))
            f.params.child('npad').setValue(pad)
            f.params.child('npad').setDefault(pad)
        elif f.func_name == 'Crop':
            pad = int(np.ceil((shape[2] * np.sqrt(2) - shape[2]) / 2))
            f.params.child('p11').setValue(pad)
            f.params.child('p11').setDefault(pad)
            f.params.child('p12').setValue(pad)
            f.params.child('p12').setDefault(pad)
            f.params.child('p21').setValue(pad)
            f.params.child('p21').setDefault(pad)
            f.params.child('p22').setValue(pad)
            f.params.child('p22').setDefault(pad)

        elif f.func_name == 'Write':  # dataset specific write values
            data_folders = {'bl832data-raw': 'bl832data-scratch', 'data-raw': 'data-scratch'}
            file_name = path.split("/")[-1].split(".")[0]
            working_dir = path.split(file_name)[0]
            for key in data_folders.keys():
                if key in working_dir:
                    user = working_dir.split('/' + key)[-1].split('/')[1]
                    mount = working_dir.split(key)[0]
                    working_dir = os.path.join(mount, data_folders[key], user)
            outname = os.path.join(working_dir, *2 * ('RECON_' + file_name,))
            f.params.child('parent folder').setValue(working_dir)
            f.params.child('parent folder').setDefault(working_dir)
            f.params.child('folder name').setValue('RECON_' + file_name)
            f.params.child('folder name').setDefault('RECON_' + file_name)
            f.params.child('file name').setValue('RECON_' + file_name)
            f.params.child('file name').setDefault('RECON_' + file_name)
            f.params.child('fname').setValue(outname)
            f.params.child('fname').setDefault(outname)
        if f.input_functions:
            set_als832_defaults(mdata, funcwidget_list=f.input_functions.values(), path=path, shape=shape)


def set_reader_defaults(reader_widget, shape, cpu):
    """
    Sets defaults for reader widget based on dataset size
    """
    reader_widget.params.child('start_sinogram').setLimits([0, shape[2]])
    reader_widget.params.child('end_sinogram').setLimits([0, shape[2]])
    reader_widget.params.child('step_sinogram').setLimits([1, shape[2] + 1])
    reader_widget.params.child('start_projection').setLimits([0, shape[0]])
    reader_widget.params.child('end_projection').setLimits([0, shape[0]])
    reader_widget.params.child('step_projection').setLimits([1, shape[0] + 1])
    reader_widget.params.child('start_width').setLimits([0, shape[1]])
    reader_widget.params.child('end_width').setLimits([0, shape[1]])
    reader_widget.params.child('step_width').setLimits([1, shape[2] + 1])

    reader_widget.params.child('end_sinogram').setValue(shape[2])
    reader_widget.params.child('end_sinogram').setDefault(shape[2])
    reader_widget.params.child('end_projection').setValue(shape[0])
    reader_widget.params.child('end_projection').setDefault(shape[0])
    reader_widget.params.child('end_width').setValue(shape[1])
    reader_widget.params.child('end_width').setDefault(shape[1])
    reader_widget.params.child('sinograms_per_chunk').setValue(cpu * 5)
    reader_widget.params.child('projections_per_chunk').setValue(cpu * 5)
    reader_widget.params.child('sinograms_per_chunk').setDefault(cpu * 5)
    reader_widget.params.child('projections_per_chunk').setDefault(cpu * 5)



def extract_pipeline_dict(funwidget_list):
    """
    Extract a dictionary from a FunctionWidget list in the appropriate format to save as a yml file

    Parameters
    ----------
    funwidget_list : list of FunctionWidgets
        list of FunctionWidgets exposed in the UI workflow pipeline

    Returns
    -------
    dict
        dictionary specifying the workflow pipeline
    """

    # list of parameter name exceptions, for the "Write" function
    ex_lst = ['file name', 'fname', 'folder name', 'parent folder']

    d = OrderedDict()
    count = 1
    for f in funwidget_list:
        # a bunch of special cases for the write function
        func_name = str(count) + ". " + f.func_name
        if "Write" in f.func_name:
            write_dict = OrderedDict()
            d[func_name] = OrderedDict({f.subfunc_name: write_dict})
            for child in f.params.children():
                if child.name() in ex_lst:
                    d[func_name][f.subfunc_name][child.name()] = str(child.value())
                elif child.name() == "Browse":
                    pass
                else:
                    d[func_name][f.subfunc_name][child.name()] = child.value()
        else:
            d[func_name] = {f.subfunc_name: {'Parameters': {p.name(): p.value() for p in f.params.children()}}}
        d[func_name][f.subfunc_name]['Enabled'] = f.enabled
        if f.func_name == 'Reconstruction':
            d[func_name][f.subfunc_name].update({'Package': f.packagename})
        for param, ipf in f.input_functions.iteritems():
            if 'Input Functions' not in d[func_name][f.subfunc_name]:
                d[func_name][f.subfunc_name]['Input Functions'] = {}
            id = {ipf.func_name: {ipf.subfunc_name: {'Parameters': {p.name(): p.value() for p in ipf.params.children()}}}}
            d[func_name][f.subfunc_name]['Input Functions'][param] = id
        count += 1
    return d

def extract_runnable_dict(funwidget_list):
    """
    Extract a dictionary from a FunctionWidget list in the appropriate format to save as a python runnable.

    Parameters
    ----------
    funwidget_list : list of FunctionWidgets
        list of FunctionWidgets exposed in the UI workflow pipeline

    Returns
    -------
    dict
        dictionary specifying the workflow pipeline and important parameters
    """
    center_functions = {'find_center_pc': {'proj1': 'tomo[0]', 'proj2': 'tomo[-1]'},
                        'find_center': {'tomo': 'tomo', 'theta': 'theta'}, 'find_center_vo': {'tomo': 'tomo'}}

    d = OrderedDict()
    func_dict = OrderedDict(); subfuncs = OrderedDict()
    count = 1
    for f in funwidget_list:
        keywords = {}
        if not f.enabled or 'Reader' in f.name:
            continue

        func = "{}.{}".format(f.package, f._function.func_name)
        if 'xicam' in func:
            func = func.split(".")[-1]
        fpartial = f.partial
        for key, val in fpartial.keywords.iteritems():
            keywords[key] = val
        for arg in inspect.getargspec(f._function)[0]:
            if arg not in f.partial.keywords.iterkeys() or 'center' in arg:
                keywords[arg] = arg


        # get rid of degenerate keyword arguments
        if 'arr' in keywords and 'tomo' in keywords:
            keywords['tomo'] = keywords['arr']
            keywords.pop('arr', None)

        # special cases for the 'write' function
        if 'start' in keywords:
            keywords['start'] = 'start'
        if 'Write' in f.name:
            keywords.pop('parent folder', None)
            keywords.pop('folder name', None)
            keywords.pop('file name', None)


        if 'Reconstruction' in f.name:
            for param, ipf in f.input_functions.iteritems():
                if 'theta' in param or 'center' in param:
                    subfunc = "{}.{}(".format(ipf.package,ipf._function.func_name)
                    for key, val in ipf.partial.keywords.iteritems():
                        subfunc += "{}={},".format(key, val) if not isinstance(val, str) \
                            else '{}=\'{}\','.format(key, val)
                    for cor_func in center_functions.iterkeys():
                        if ipf._function.func_name == cor_func:
                            for k, v in center_functions[cor_func].iteritems():
                                subfunc += "{}={},".format(k, v)
                    subfunc += ")"
                    subfuncs[param] = subfunc
            if 'astra' in keywords['algorithm']:
                keywords['algorithm'] = 'tomopy.astra'

        func_dict[str(count) + ". " + func] = keywords
        count += 1

    d['func'] = func_dict
    d['subfunc'] = subfuncs
    return d




__author__ = "Luis Barroso-Luque, Holden Parks"
__copyright__ = "Copyright 2016, CAMERA, LBL, ALS"
__credits__ = ["Ronald J Pandolfi", "Dinesh Kumar", "Singanallur Venkatakrishnan", "Luis Luque",
               "Holden Parks", "Alexander Hexemer"]
__license__ = ""
__version__ = "1.2.1"
__maintainer__ = "Ronald J Pandolfi"
__email__ = "ronpandolfi@lbl.gov"
__status__ = "Beta"

from tomocam import tomoCam

def recon(tomo, theta, center=None, algorithm=None, input_params=None, **kwargs):
    """
    Wrapper for TomoCam reconstruction functions

    Parameters
    ----------
    tomo : np.ndarray
        3D tomography data
    theta : list
        List of angles at which tomography data was taken
    center : float, optional
        Center of rotation of dataset
    algorithm : str, optional
        Name of algorithm to be run (either 'gridrec' or 'sirt'
    input_params : dict, optional
        Dictionary of other params for reconstruction
    """

    if 'gridrec' in algorithm:
        return tomoCam.gpuGridrec(tomo, theta, center, input_params)
    elif 'sirt' in algorithm:
        return tomoCam.gpuSIRT(tomo, theta, center, input_params)
    else:
        raise ValueError('TomoCam reconstruction must be either \'gridrec\', \'sirt\', or \'mbir\'')



import collections
import numpy as np
import loader
from scipy import signal
import integration
import writer
#from skimage.measure import block_reduce  # Use this to subsample if you want

# README!
#
# To add a variation operation,
# 1. Your function must adhere to the signature below:
# (data, t, roi)
#    Where data is a list of ndarray thumbnails (5x), one for each frame,
#    t is the frame index to evaluate at, and roi is the region of interest mask.
# 2. Add your function to the 'operations' dictionary at the end of this module. The key is the display name.


# Operation signature:
# If calculating variation between only two consecutive frames, use current and previous
# ROI-mask is accept=1

def chisquared(data, t, roi):
    current = data[t].astype(float)
    previous = data[t - 1].astype(float)
    return np.sum(roi * np.square(current - previous))

def imgmax(data, t, roi):
    current = data[t].astype(float)
    return (roi * current).max()


def qmax(data, t, roi):
    current = data[t]
    q, I = integration.radialintegratepyFAI(current, cut=roi)[:2]
    return q[np.argmax(I)]


def fit_mean(data, t, roi):
    current = data[t]
    q, I = integration.radialintegratepyFAI(current, cut=roi)[:2]
    from astropy.modeling import models, fitting
    qmin = q[(np.array(I) != 0).argmax()]
    qmax = q[len(I) - 1 - (np.array(I) != 0)[::-1].argmax()]
    g_init = models.Gaussian1D(amplitude=np.array(I).max(), mean=(qmax + qmin) / 2, stddev=(qmax - qmin) / 4)
    fit_g = fitting.LevMarLSQFitter()
    g = fit_g(g_init, q, I)
    return g.mean.value


def fit_stddev(data, t, roi):
    current = data[t]
    q, I = integration.radialintegratepyFAI(current, cut=roi)[:2]
    from astropy.modeling import models, fitting
    qmin = q[(np.array(I) != 0).argmax()]
    qmax = q[len(I) - 1 - (np.array(I) != 0)[::-1].argmax()]
    g_init = models.Gaussian1D(amplitude=np.array(I).max(), mean=(qmax + qmin) / 2, stddev=(qmax - qmin) / 4)
    fit_g = fitting.LevMarLSQFitter()
    g = fit_g(g_init, q, I)
    return g.stddev.value


def gaussian_fit(data, t, roi):
    current = data[t]
    q, I = integration.radialintegratepyFAI(current, cut=roi)[:2]
    from astropy.modeling import models, fitting
    qmin = q[(np.array(I) != 0).argmax()]
    qmax = q[len(I) - 1 - (np.array(I) != 0)[::-1].argmax()]
    g_init = models.Gaussian1D(amplitude=np.array(I).max(), mean=(qmax + qmin) / 2, stddev=(qmax - qmin) / 4)
    fit_g = fitting.LevMarLSQFitter()
    g = fit_g(g_init, q, I)
    return collections.OrderedDict(
        [('stddev', g.stddev.value), ('mean', g.mean.value), ('Amplitude', g.amplitude.value),
         ('chi^2', np.average((g(q) - I) ** 2))])

def absdiff(data, t, roi):
    current = data[t].astype(float)
    previous = data[t - 1].astype(float)
    return np.sum(roi * np.abs(current - previous))


def normabsdiff(data, t, roi):
    current = data[t].astype(float)
    previous = data[t - 1].astype(float)
    return np.sum(roi * np.abs(current - previous) / previous)


def sumintensity(data, t, roi):
    current = data[t].astype(float)
    return np.sum(roi * current)


def normabsdiffderiv(data, t, roi):
    current = data[t].astype(float)
    previous = data[t - 1].astype(float)
    next = data[t + 1].astype(float)
    return -np.sum(roi * (np.abs(next - current) / current) + np.sum(np.abs(current - previous) / current))


def chisquaredwithfirst(data, t, roi):
    current = data[t].astype(float)
    first = data[0].astype(float)
    return np.sum(roi * np.square(current.astype(float) - first))


def radialintegration(data, t, roi):
    current = data[t]
    return integration.radialintegratepyFAI(current, cut=roi)[:2]


def angularcorrelationwithfirst(data, t, roi):
    # ROI is assumed to be in cake mode

    experiment.center = (experiment.center[0] / 5, experiment.center[1] / 5)

    currentcake, _, _ = integration.cake(data[t], experiment)
    firstcake, _, _ = integration.cake(data[0], experiment)
    # cakeroi, _, _ = integration.cake(np.ones_like(data), experiment)

    currentchi = np.sum(currentcake * roi, axis=0)
    firstchi = np.sum(firstcake * roi, axis=0)

    return signal.convolve(currentchi, firstchi)




operations = collections.OrderedDict([('Chi Squared', chisquared),
                                      ('Max', imgmax),
                                      ('Absolute Diff.', absdiff),
                                      ('Norm. Abs. Diff.', normabsdiff),
                                      ('Sum Intensity', sumintensity),
                                      ('Norm. Abs. Derivative', normabsdiffderiv),
                                      ('Chi Squared w/First Frame', chisquaredwithfirst),
                                      ('Angular autocorrelation w/First Frame', angularcorrelationwithfirst),
                                      ('Radial Integration', radialintegration),
                                      ('Q at peak max', qmax),
                                      ('Q at peak fit max', fit_mean),
                                      ('Peak Fit Std. Dev.', fit_stddev),
                                      ('Gaussian Q Fit', gaussian_fit)
                                      ])


experiment = None


# --coding: utf-8 --
#! /usr/bin/env python

from collections import OrderedDict
from xicam import debugtools


################ Rules ###############
# Rule failure represents an exclusion

iseven = lambda m: not(m % 2)
is6multiple = lambda m: not(m%6) and m!=0       # TODO: check that the m!=0 condition is correct
is4multiple = lambda m: not(m%4) and m!=0
is3multiple = lambda m: not(m%3) and m!=0
is2multiple = lambda m: not(m%2) and m!=0

h = lambda h,k,l: iseven(h)
k = lambda h,k,l: iseven(k)
l = lambda h,k,l: iseven(l)
hpk = lambda h,k,l: iseven(h+k)
hpl = lambda h,k,l: iseven(h+l)
kpl = lambda h,k,l: iseven(k+l)
hpkpl = lambda h,k,l: iseven(h+k+l)
kpl4n = lambda h,k,l: is4multiple(k+l)
hpl4n = lambda h,k,l: is4multiple(h+l)
hpk4n = lambda h,k,l: is4multiple(h+k)
h4n = lambda h,k,l: is4multiple(h)
k4n = lambda h,k,l: is4multiple(k)
l4n = lambda h,k,l: is4multiple(l)
twohpl4n = lambda h,k,l: is4multiple(2*h+l)
l3n = lambda h,k,l: is3multiple(l)
l6n = lambda h,k,l: is6multiple(l)
mhpkpl3n = lambda h,k,l: is3multiple(-h+k+l)
hpl3n = lambda h,k,l: is3multiple(h+l)
hmkpl3n = lambda h,k,l: is3multiple(h-k+l)
mhpl3n = lambda h,k,l: is3multiple(-h+l)
h2n = lambda h,k,l: is2multiple(h)
k2n = lambda h,k,l: is2multiple(k)
l2n = lambda h,k,l: is2multiple(l)
kpl2n = lambda h,k,l: is2multiple(k+l)




######################################

cache = dict()


def arezeros(*args):
    if len(args)==4:
        zh,zk,zl=args[:3]
        mh,mk,ml=args[3]
        if (bool(mh) and zh) or (bool(mk) and zk) or (bool(ml) and zl):
            return False
        return True

    elif len(args)==5:
        zh,zk,zi,zl=args[:4]
        mh,mk,mi,ml=args[4]
        if (bool(mh) and zh) or (bool(mk) and zk) or (bool(mi) and zi) or (bool(ml) and zl):
            return False
        return True


class SGClass:
    conditions = OrderedDict([])

    def check(self,m,SG):
        """
        Get the relevant exclusion rule column and then test against it.
        """
        col,m=self.getcolumn(m)
        if type(self.conditions[SG]) is not list: # ALLOW LINKED CONDITIONS
            SG = self.conditions[SG]
        return self.checkcolumn(SG,col,m)

    def getcolumn(self,m):
        """
        Determine which column of the exclusion table to test with given miller indices 'm'.
        Unique SGClass subclasses may override this if there are more columns or the column conditions are unique.
        This virtual class is overridden by each crystal system.
        """
        pass

    def checkcolumn(self,SG,col,m):
        """
        Check the exclusion rules in column 'col' with miller indices 'm'.
        """
        if col is None:         # None means no column applies.
            return True
        if len(m)==4:
            m=(m[0],m[1],m[3])

        SGconditions = self.conditions[SG][col]
        if type(SGconditions) is not list:
            SGconditions = [SGconditions]
        for condition in SGconditions:
            if condition is not None:
                if not condition(*m):   # Do not change this to 'return not condition(*m)'!
                    return False
        return True

class Triclinic(SGClass):
    conditions = OrderedDict([('P1',[]),
                              (u'P1\u0305',[])])
    def check(self,m,SG):
        return True

class Monoclinic(SGClass):
    ############## IMPORTANT NOTE ################
    # Monoclinic must have the unique axis as b for this to work!
    # Status: COMPLETE!
    conditions = OrderedDict([('P2',[None,None,None]), #3
                              (u'P2₁',[None,None,k]), #4
                              ('C2',[hpk,h,k]), #5
                              ('A2',[kpl,l,k]),
                              ('I2',[hpkpl,hpl,k]),
                              ('Pm',[None,None,None]),#6
                              ('Pc',[None,l,None]),#7
                              ('Pa',[None,h,None]),
                              ('Pn',[None,hpl,None]),
                              ('Cm',[hpk,h,k]),#8
                              ('Am',[hpk,h,k]),
                              ('Im',[hpk,h,k]),
                              ('Cc',[hpk,[h,l],k]),#9
                              ('An',[kpl,[h,l],k]),
                              ('Ia',[hpkpl,[h,l],k]),
                              ('P2/m',[None,None,None]),#10
                              (u'P2₁/m',[None,None,k]),#11
                              ('C2/m',[hpk,h,k]),#12
                              ('A2/m',[kpl,l,k]),
                              ('I2/m',[hpkpl,hpl,k]),
                              ('P2/c',[None,l,None]),#13
                              ('P2/a',[None,h,None]),
                              ('P2/n',[None,hpl,None]),
                              (u'P2₁/c',[None,l,k]),#14
                              (u'P2₁/a',[None,h,k]),
                              (u'P2₁/n',[None,hpl,k]),
                              ('C2/c',[hpk,[h,l],k]),#15
                              ('A2/n',[kpl,[h,l],k]),
                              ('I2/a',[hpkpl,[h,l],k])])



    def getcolumn(self,m):
        if arezeros(1,0,1,m):
            column = 2
        elif arezeros(0,1,0,m) or arezeros(1,1,0,m) or arezeros(0,1,1,m):
            column = 1
        elif arezeros(0,0,0,m) or arezeros(1,0,0,m) or arezeros(0,0,1,m):   # TODO: why doesn't this column just say hkl?
            column = 0
        else:
            debugtools.frustration()
            raise ValueError
        return column,m

class Orthorhombic(SGClass):
    # Status: COMPLETE!
    conditions = OrderedDict([('P222',[None,None,None,None,None,None,None]), #16
                              (u'P222₁',[None,None,None,None,None,None,l]),#17
                              (u'P22₁2',[None,None,None,None,None,k,None]),
                              (u'P2₁22',[None,None,None,None,h,None,None]),
                              (u'P2₁2₁2',[None,None,None,None,h,k,None]),#18
                              (u'P22₁2₁',[None,None,None,None,None,k,l]),
                              (u'P2₁22₁',[None,None,None,None,h,None,l]),
                              (u'P2₁2₁2₁',[None,None,None,None,h,k,l]),#19
                              (u'C222₁',[hpk,k,h,hpk,h,k,l]),#20,unfilled to 25
                              ('C222',[hpk,k,h,hpk,h,k,None]),#21
                              ('F222',[[hpk,hpl,kpl],[k,l],[h,l],[h,k],h,k,l]),#22
                              ('I222',[hpkpl,kpl,hpl,hpk,h,k,l]),#23
                              (u'I2₁2₁2₁',[hpkpl,kpl,hpl,hpk,h,k,l]),#24
                              ('Pmm2',[None,None,None,None,None,None,None]),#25
                              ('Pm2m',[None,None,None,None,None,None,None]),
                              ('P2mm',[None,None,None,None,None,None,None]),
                              (u'Pmc2₁',[None,None,l,None,None,None,l]),#26
                              (u'P2₁ma',[None,None,None,h,h,None,None]),
                              (u'Pm2₁b',[None,None,None,k,None,k,None]),
                              (u'P2₁am',[None,None,h,None,h,None,None]),
                              ('Pcc2',[None,l,l,None,None,None,l]),#27
                              ('P2aa',[None,None,h,h,h,None,None]),
                              ('Pma2',[None,None,h,None,h,None,None]),#28
                              ('Pm2a',[None,None,None,h,h,None,None]),
                              ('P2mb',[None,None,None,h,None,k,None]),
                              ('P2cm',[None,None,l,None,None,None,l]),
                              (u'Pca2₁',[None,l,h,None,h,None,l]),#29
                              (u'P2₁ab',[None,None,h,k,h,k,None]),
                              (u'P2₁ca',[None,None,h,None,h,None,l]),
                              ('Pnc2',[None,kpl,l,None,None,k,l]),
                              (u'Pmn2₁',[None,None,hpl,None,h,None,l]), #31
                              (u'Pmn2₁',[None,None,hpl,None,h,None,l]),
                              (u'P2₁nm', [None, None, None, None, None, None, None]),
                              (u'Pmnm', [None, None, None, None, None, None, None]),
                              (u'P2na', [None, None, hpl, h, h, None, l]),
                              (u'Pmna', [None, None, hpl, h, h, None, l]),
                              (u'P2na', [None, None, hpl, k, h, k, l]),
                              (u'P2nm', [None, None, hpl, hpk, h, k, l]),
                              (u'Pmnn', [None, None, hpl, hpk, h, k, l]),
                              (u'Pbm2', [None, k,None,None,None,k,None]),
                              (u'Pb2₁m', [None, None, None, None, None, None, None]),
                              (u'Pbmm', [None, None, None, None, None, None, None]),
                              (u'Pb2b', [None, k, None, k, None, k, None]),
                              (u'Pbmb', [None, k, None, k, None, k, None]),
                              (u'Pb2n', [None, k, None, hpk, h, k, None]),
                              (u'Pbmn', [None, k, None, hpk, h, k, None]),
                              (u'Pba2', [None, k, h, None, k, h, None]),
                              (u'Pbam', [None, k, h, None, k, h, None]),
                              (u'Paa',  [None, k, h, h, h, k, None]),
                              (u'Pbab', [None, k, h, k, h, k, None]),
                              (u'Pan',  [None, k, h, hpk, h, k, None]),
                              (u'Pbc2₁', [None, k, l, None, None, k, l]),
                              (u'Pbcm',  [None, k, l, None, None, k, l]),
                              (u'Pbca',  [None, k, l, h, h, k, l]),
                              (u'Pbcb',  [None,k, l, k, None, k, l]),
                              (u'Pbcn',  [None, k, l, hpk, h, k, l]),
                              (u'Pbn2₁', [None, k, hpl, None, h, k, l]),
                              (u'Pbnm',  [None, k, hpl, None, h, k, l]),
                              (u'Pbna', [None, k, hpl, h, h, k, l]),
                              (u'Pbnb', [None, k, hpl, k, h, k, l]),
                              (u'Pbnn', [None, k, hpl, hpk, h, k, l]),
                              (u'Pcm2₁', [None, l, None, None, None, None, l]),
                              (u'Pc2m', [None, None, None, None, None, None, None]),
                              (u'Pcmm', [None, None, None, None, None, None, None]),
                              (u'Pc2a', [None, l, None, h, h, None, l]),
                              (u'Pcma', [None, l, None, h, h, None, l]),
                              (u'Pc2₁b', [None, l, None, k, None, k, l]),
                              (u'Pcmb',  [None, l, None, k, None, k, l]),
                              (u'Pc2₁n', [None, l, None, hpk, h, k, l]),
                              (u'Pcmn',  [None, l, None, hpk, h, k, l]),
                              (u'Pca2₁', [None, l, h, None, h, None, l]),
                              (u'Pcam',  [None, l, h, None, h, None, l]),
                              (u'Pcaa', [None, l, h, h, h, None, l]),
                              (u'Pcab', [None, l, h, k, h, k, l]),
                              (u'Pcan', [None, l, h, hpk, h, k, l]),
                              (u'Pcc2', [None, l, h, None, None, None, l]),
                              (u'Pccm', [None, l, l, None, None, None, l]),
                              (u'Pcca', [None, l, l, h, h, None, l]),
                              (u'Pccb', [None, l, l, k, None, k, l]),
                              (u'Pccn', [None, l, l, hpk, h, k, l]),
                              (u'Pcn2', [None, l, hpl, None, h, None, l]),
                              (u'Pcnm', [None, l, hpl, None, h, None, l]),
                              (u'Pcna', [None, l, hpl, h, h, None, l]),
                              (u'Pcnb', [None, l, hpl, k, h, k, l]),
                              (u'Pcnn', [None, l, hpl, hpk, h, k, l]),
                              (u'Pnm2₁', [None, kpl, None, None, None, k, l]),
                              (u'Pnmm', [None, kpl, None, None, None, k, l]),
                              (u'Pn2₁m', [None, None, None, None, None, None, None]),
                              (u'Pn2₁a', [None, kpl, None, h, h, k, l]),
                              (u'Pnma',  [None, kpl, None, h, h, k, l]),
                              (u'Pn2b',  [None, kpl, None, k, None, k, l]),
                              (u'Pnmb',  [None, kpl, None, k, None, k, l]),
                              (u'Pn2n', [None, kpl, None, hpk, h, k, l]),
                              (u'Pnmn', [None, kpl, None, hpk, h, k, l]),
                              (u'Pna2₁', [None, kpl, h, None, h, k, l]),
                              (u'Pnam',  [None, kpl, h, None, h, k, l]),
                              (u'Pnaa', [None, kpl, h, h, h, k, l]),
                              (u'Pnab', [None, kpl, h, k, h, k, l]),
                              (u'Pnan', [None, kpl, h, hpk, h, k, l]),
                              (u'Pnc2', [None, kpl, l, None, None, k, l]),
                              (u'Pncm', [None, kpl, l, None, None, k, l]),
                              (u'Pnca', [None, kpl, l, h, h, k, l]),
                              ('Pncb', [None,kpl,l,k,None,k,l]),
                              ('Pncn', [None, kpl, l,hpk,h,k,l]),
                              ('Pnn2', [None, kpl, hpl,None,h,k,l]),
                              ('Pnnm', [None, kpl, hpl,None,h,k,l]),
                              ('Pnna', [None, kpl, hpl,h,h,k,l]),
                              ('Pnnb', [None, kpl, hpl,k,h,k,l]),
                              ('Pnnn', [None, kpl, hpl,hpk,h,k,l]),
                              ('C222', [hpk, k, h,hpk,h,k,None]),
                              ('Cmm2', [hpk, k, h,hpk,h,k,None]),
                              ('Cmmm', [hpk, k, h,hpk,h,k,None]),
                              ('Cm2m', [hpk, k, h,hpk,h,k,None]),
                              ('C2mm', [hpk, k, h,hpk,h,k,None]),
                              (u'C222\u2081',[hpk, k, h,hpk,h,k,l]),
                              ('Cm2e', [hpk, k, h,[h,k],h,k,None]),
                              ('Cmme', [hpk, k, h,[h,k],h,k,None]),
                              ('C2me', [hpk, k, h,[h,k],h,k,None]),
                              (u'Cm2\u2081', [hpk, k, [h,l],hpk,h,k,l]),
                              ('Cmcm', [hpk, k, [h,l],hpk,h,k,l]),
                              ('C2cm', [hpk, k, [h,l],hpk,h,k,l]),
                              ('C2ce', [hpk,k,[h,l],[h,k],h,k,l]),
                              ('Cmce',   [hpk,k,[h,l],[h,k],h,k,l]),
                              (u'Ccm2\u2081', [hpk,[k,l],h,hpk,h,k,l]),
                              ('Ccmm', [hpk,[k,l],h,hpk,h,k,l]),
                              ('Cc2m', [hpk,[k,l],h,hpk,h,k,l]),
                              ('Cc2e', [hpk,[k,l],h,[h,k],h,k,l]),
                              ('Ccme', [hpk,[k,l],h,[h,k],h,k,l]),
                              ('Ccc2', [hpk,[k,l],[h,l],hpk,h,k,l]),
                              ('Cccm', [hpk,[k,l],[h,l],hpk,h,k,l]),
                              ('Ccce', [hpk,[k,l],[h,l],[h,k],h,k,l]),
                              ('B222', [hpl,l,hpl,h,h,None,l]),
                              ('Bmm2', [hpl,l,hpl,h,h,None,l]),
                              ('Bmmm', [hpl,l,hpl,h,h,None,l]),
                              ('Bm2m', [hpl,l,hpl,h,h,None,l]),
                              ('B2mm',   [hpl,l,hpl,h,h,None,l]),
                              (u'B22\u20812', [hpl,l,hpl,h,h,k,l]),
                              (u'Bm2\u2081b', [hpl,l,hpl,[h,k],h,k,l]),
                              ('Bmmb', [hpl,l,hpl,[h,k],h,k,l]),
                              ('B2mb', [hpl,l,hpl,[h,k],h,k,l]),
                              ('Bm2e', [hpl,l,[h,l],h,h,None,l]),
                              ('Bmem', [hpl,l,[h,l],h,h,None,l]),
                              ('B2em', [hpl,l,[h,l],h,h,None,l]),
                              ('B2eb', [hpl,l,[h,l],[h,k],h,k,l]),
                              ('Bmeb', [hpl,l,[h,l],[h,k],h,k,l]),
                              ('Bbm2', [hpl,[k,l],hpl,h,h,k,l]),
                              ('Bbmm', [hpl,[k,l],hpl,h,h,k,l]),
                              (u'Bb2\u2081m', [hpl,[k,l],hpl,h,h,k,l]),
                              ('Bb2b', [hpl,[k,l],hpl,[h,k],h,k,l]),
                              ('Bbmb', [hpl,[k,l],hpl,[h,k],h,k,l]),
                              ('Bbe2', [hpl,[k,l],[h,l],h,h,k,l]),
                              ('Bbem', [hpl,[k,l],[h,l],h,h,k,l]),
                              ('Bbeb', [hpl,[k,l],[h,l],[h,k],h,k,l]),
                              ('A222', [kpl,kpl,l,k,None,k,l]),
                              ('Amm2', [kpl,kpl,l,k,None,k,l]),
                              ('Ammm', [kpl,kpl,l,k,None,k,l]),
                              ('Am2m', [kpl,kpl,l,k,None,k,l]),
                              ('A2mm', [kpl,kpl,l,k,None,k,l]),
                              (u'A2\u208122', [kpl,kpl,l,k,h,k,l]),
                              ('Am2a', [kpl,kpl,l,[h,k],h,k,l]),
                              ('Amma', [kpl,kpl,l,[h,k],h,k,l]),
                              (u'A2\u2081ma', [kpl,kpl,l,[h,k],h,k,l]),
                              ('Ama2', [kpl,kpl,[h,l],k,h,k,l]),
                              ('Amam', [kpl,kpl,[h,l],k,h,k,l]),
                              ('A2aa', [kpl,kpl,[h,l],[h,k],h,k,l]),
                              ('Amaa', [kpl,kpl,[h,l],[h,k],h,k,l]),
                              ('Aem2', [kpl,[k,l],l,k,None,k,l]),
                              ('Aemm', [kpl,[k,l],l,k,None,k,l]),
                              ('Ae2m', [kpl,[k,l],l,k,None,k,l]),
                              ('Ae2a', [kpl,[k,l],l,[h,k],h,k,l]),
                              ('Aema', [kpl,[k,l],l,[h,k],h,k,l]),
                              ('Aea2', [kpl,[k,l],[h,l],k,h,k,l]),
                              ('Aeam', [kpl,[k,l],[h,l],k,h,k,l]),
                              ('Aeaa', [kpl,[k,l],[h,l],[h,k],h,k,l]),
                              ('I222', [hpkpl,kpl,hpl,hpk,h,k,l]),
                              ('Imm2', [hpkpl,kpl,hpl,hpk,h,k,l]),
                              ('Immm', [hpkpl,kpl,hpl,hpk,h,k,l]),
                              (u'I2\u20812\u20812\u2081', [hpkpl,kpl,hpl,hpk,h,k,l]),
                              ('Im2m', [hpkpl,kpl,hpl,hpk,h,k,l]),
                              ('F222', [[hpk,hpl,kpl],[k,l],[h,l],[h,k],h,k,l]),
                              ('I2mm', [hpkpl,kpl,hpl,hpk,h,k,l]),
                              ('Im2a',[hpkpl,kpl,hpl,[h,k],h,k,l]),
                              ('I2mb',[hpkpl,kpl,hpl,[h,k],h,k,l]),
                              ('Ima2',[hpkpl,kpl,[h,l],hpk,h,k,l]),
                              ('I2cm',[hpkpl,kpl,[h,l],hpk,h,k,l]),
                              ('I2cb',[hpkpl,kpl,[h,l],[h,k],h,k,l]),
                              ('Iem2',[hpkpl,kpl,[h,l],[h,k],h,k,l]),
                              ('Ic2a',[hpkpl,[k,l],hpl,[h,k],h,k,l]),
                              ('Iba2',[hpkpl,[k,l],[h,l],hpk,h,k,l]),
                              ('Fmm2',[[hpk,hpl,kpl],[k,l],[h,l],[h,k],h,k,l]),
                              ('Fm2m',[[hpk,hpl,kpl],[k,l],[h,l],[h,k],h,k,l]),
                              ('F2mm',[[hpk,hpl,kpl],[k,l],[h,l],[h,k],h,k,l]),
                              ('F2dd',[[hpk,hpl,kpl],[k,l],[hpl4n,h,l],[hpk4n,h,k],h4n,k4n,l4n]),
                              ('Fd2d',[[hpk,hpl,kpl],[kpl4n,k,l],[h,l],[hpk4n,h,k],h4n,k4n,l4n]),
                              ('Fdd2',[[hpk,hpl,kpl],[kpl4n,k,l],[hpl4n,h,l],[h,k],h4n,k4n,l4n]),
                              ('Imma',[hpkpl,kpl,hpl,[h,k],h,k,l]),
                              ('Immb',[hpkpl,kpl,hpl,[h,k],h,k,l]),
                              ('Imam',[hpkpl,kpl,[h,l],hpk,h,k,l]),
                              ('Imcm',[hpkpl,kpl,[h,l],hpk,h,k,l]),
                              ('Imcb',[hpkpl,kpl,[h,l],[h,k],h,k,l]),
                              ('Iemm',[hpkpl,kpl,[h,l],[h,k],h,k,l]),
                              ('Icma',[hpkpl,[k,l],hpl,[h,k],h,k,l]),
                              ('Ibam',[hpkpl,[k,l],[h,l],hpk,h,k,l]),
                              ('Ibca',[hpkpl,[k,l],[h,l],[h,k],h,k,l]),
                              ('Icab',[hpkpl,[k,l],[h,l],[h,k],h,k,l]),
                              ('Fmmm',[[hpk,hpl,kpl],[k,l],[h,l],[h,k],h,k,l]),
                              ('Fddd',[[hpk,hpl,kpl],[kpl4n,k,l],[hpl4n,h,l],[hpk4n,h,k],h4n,k4n,l4n])
    ])

    def getcolumn(self,m):
        if arezeros(1,1,0,m):
            column = 6
        elif arezeros(1,0,1,m):
            column = 5
        elif arezeros(0,1,1,m):
            column = 4
        elif arezeros(0,0,1,m):
            column = 3
        elif arezeros(0,1,0,m):
            column = 2
        elif arezeros(1,0,0,m):
            column = 1
        else:
            column = 0
        return column,m

class Tetragonal(SGClass):
    #Status: COMPLETE!

    conditions = OrderedDict([('P4',[None,None,None,None,None,None,None]),
                              (u'P4\u0305','P4'),
                              ('P4/m','P4'),
                              ('P422','P4'),
                              ('P4mm','P4'),
                              (u'P4\u03052m','P4'),
                              ('P4/mmm','P4'),
                              (u'P4\u0305m2','P4'),
                              (u'P42₁2',[None,None,None,None,None,k,None]),
                              (u'P4\u03052₁m',u'P42₁2'),
                              (u'P4₂',[None,None,None,None,l,None,None]),
                              (u'P4₂/m',u'P4₂'),
                              (u'P4₂22',u'P4₂'),
                              (u'P4₂2₁2',[None,None,None,None,l,k,None]),
                              (u'P4₁',[None,None,None,None,l4n,None,None]),
                              (u'P4₃',u'P4₁'),
                              (u'P4₁22',u'P4₁'),
                              (u'P4₃22',u'P4₁'),
                              (u'P4₁2₁2',[None,None,None,None,l4n,k,None]),
                              (u'P4₃2₁2',u'P4₁2₁2'),
                              (u'P4₂mc',[None,None,None,l,l,None,None]),
                              (u'P4\u03052c',u'P4₂mc'),
                              (u'P42/mmc',u'P4₂mc'),
                              (u'P4\u03052₁c',[None,None,None,l,l,k,None]),
                              ('P4bm',[None,None,k,None,None,k,None]),
                              (u'P4\u0305b2','P4bm'),
                              (u'P4/mbm','P4bm'),
                              (u'P4₂bc',[None,None,k,l,l,k,None]),
                              (u'P4₂/mbc',u'P4₂bc'),
                              (u'P4₂cm',[None,None,l,None,l,None,None]),
                              (u'P4\u0305c2',u'P4₂cm'),
                              (u'P4₂/mcm',u'P4₂cm'),
                              (u'P4cc',[None,None,l,l,l,None,None]),
                              (u'P4/mcc',u'P4cc'),
                              (u'P4₂nm',[None,None,kpl,None,l,k,None]),
                              (u'P4\u0305n2',u'P4₂nm'),
                              (u'P4₂/mnm',u'P4₂nm'),
                              (u'P4nc',[None,None,kpl,l,l,k,None]),
                              (u'P4/mnc',u'P4nc'),
                              (u'P4/n',[None,hpk,None,None,None,k,None]),
                              (u'P4/nmm',u'P4/n'),
                              (u'P4₂/n',[None,hpk,None,None,l,k]),
                              (u'P4₂/nmc',[None,hpk,None,l,l,k,None]),
                              (u'P4/nbm',[None,hpk,k,None,None,k,None]),
                              (u'P4₂/nbc',[None,hpk,k,l,l,k,None]),
                              (u'P4₂/ncm',[None,hpk,l,None,l,k,None]),
                              (u'P4/ncc',[None,hpk,l,l,l,k,None]),
                              (u'P4₂nnm',[None,hpk,kpl,None,l,k,None]),
                              (u'P4/nnc',[None,hpk,kpl,l,l,k,None]),
                              (u'I4',[hpkpl,hpk,kpl,l,l,k,None]),
                              (u'I4\u0305',u'I4'),
                              (u'I4/m',u'I4'),
                              (u'I422',u'I4'),
                              (u'I4mm',u'I4'),
                              (u'I4\u03052m',u'I4'),
                              (u'I4/mmm',u'I4'),
                              (u'I4\u0305m2',u'I4'),
                              (u'I4₁',[hpkpl,hpk,kpl,l,l4n,k,None]),
                              (u'I4₁22',u'I4₁'),
                              (u'I4₁md',[hpkpl,hpk,kpl,twohpl4n,l4n,k,h]),
                              (u'I4\u03052d',u'I4₁md'),
                              (u'I4cm',[hpk,l,hpk,[k,l],l,l,k,None]),
                              (u'I4\u0305c2',u'I4cm'),
                              (u'I4/mcm',u'I4cm'),
                              (u'I4₁cd',[hpkpl,hpk,[k,l],twohpl4n,l4n,k,h]),
                              (u'I4₁/a',[hpkpl,[h,k],kpl,l,l4n,k,None]),
                              (u'I4₁/amd',[hpkpl,[h,k],kpl,twohpl4n,l4n,k,h]),
                              (u'I4₁/acd',[hpkpl,[h,k],[k,l],twohpl4n,l4n,k,h])
                              ])

    def getcolumn(self,m):
        mh,mk,ml=m
        if mh==mk and arezeros(0,0,1,m):
            column = 6
        elif arezeros(1,0,1,m):
            column = 5
        elif arezeros(1,1,0,m):
            column = 4
        elif mh==mk:          # TODO: is is possible that no column condition is satisfied (i.e. 100). What is supposed to happen then? if the intention would be to apply column 0, why does monoclinic have hkl 0kl h0l in column 0?
            column = 3
        elif arezeros(1,0,0,m):
            column = 2
        elif arezeros(0,0,1,m):
            column = 1
        else:
            column = 0
        return column,m

class Trigonal(SGClass):
    # Status: COMPLETE!
    ## Hexagonal miller axes disabled to eliminate conflicts
    #
    # conditions = OrderedDict([('P3',[None, None, None, None]),
    #                             (u'P3\u0305',[None, None, None, None]),
    #                             ('P321',[None, None, None, None]),
    #                             ('P3m1',[None, None, None, None]),
    #                             (u'P3m\u03051',[None, None, None, None]),
    #                             ('P312',[None, None, None, None]),
    #                             ('P31m',[None, None, None, None]),
    #                             (u'P3\u03051m',[None, None, None, None]),
    #                             (u'p3₁',[None, None, None, l3n]),
    #                             (u'p3₁21',[None, None, None, l3n]),
    #                             ('p3₁12',[None, None, None, l3n]),
    #                             ('p3₂',[None, None, None, l3n]),
    #                             (u'p3₂21',[None, None, None, l3n]),
    #                             (u'p3₂12',[None, None, None, l3n]),
    #                             ('P31c',[None, None, l, l]),
    #                             (u'P3\u03051c',[None, None, l, l]),
    #                             ('P3c1',[None, l, None, l]),
    #                             (u'P3\u0305c1',[None, l, None, l]),
    #                             ('R3',[mhpkpl3n, hpl3n, l3n, l3n]),     # TODO: Why are there duplicate entries in the SG table for the R3's?
    #                             (u'R3\u0305',[mhpkpl3n, hpl3n, l3n, l3n]),
    #                             ('R32',[mhpkpl3n, hpl3n, l3n, l3n]),
    #                             ('R3m',[mhpkpl3n, hpl3n, l3n, l3n]),
    #                             (u'R3\u0305m',[mhpkpl3n, hpl3n, l3n, l3n]),
    #                             ('R3c',[mhpkpl3n, [hpl3n,l], l3n, l6n]),
    #                             (u'R3\u0305c',[mhpkpl3n, [hpl3n,l], l3n, l6n]),
    #                             ('R3c',[mhpkpl3n, [hpl3n,l], l3n, l6n]),
    #                           ])

    conditions = OrderedDict([('R3',[None,None,None,None]),
                              (u'R3\u0305','R3'),
                              (u'R32','R3'),
                              (u'R3m','R3'),
                              (u'R3\u0305m','R3'),
                              ('R3c',[None,None,l,h]),
                              (u'R3\u0305c','R3c')])

    def getcolumn(self,m):
        if len(m)==4:
            mh,mk,mi,ml = m
            if arezeros(1,1,1,0,m):
                column = 3
            elif mh==mk and mi==-2*mh:
                column = 2
            elif mh==-mk and arezeros(0,0,1,0,m):
                column = 1
            else:
                column = 0
        elif len(m)==3:
            mh,mk,ml = m
            if mh==mk==ml:
                column = 3
            elif mh==mk:
                column = 2
            else:
                column = 1
        else:
            debugtools.frustration()
            raise ValueError
        return column,m

class Hexagonal(SGClass):
    # Status: COMPLETE!
    conditions = OrderedDict([('P6',[None, None, None]),
                                (u'P6\u0305',[None, None, None]),
                                ('P6/m',[None, None, None]),
                                ('P622',[None, None, None]),
                                ('P6mm',[None, None, None]),
                                (u'P6\u03052m',[None, None, None]),
                                ('P6/mmm',[None, None, None]),
                                (u'P6\u0305m2',[None, None, None]),
                                ('P6₃', [None, None, l]),
                                ('P6₃/m', [None, None, l]),
                                ('P6₃22', [None, None, l]),
                                ('P6₂',[None, None, l3n]),
                                ('P6₂22',[None, None, l3n]),
                                ('P6₄',[None, None, None, l3n]),
                                ('P6₁',[None, None, l6n]),
                                ('P6₁22',[None, None, l6n]),
                                ('P6₃',[None, None, l6n]),
                                ('P6₅22',[None, None, l6n]),
                                ('P6₃mc',[None, l, l]),
                                (u'P6\u03052c',[None, l, l]),
                                ('P6₃/mmc',[None, l, l]),
                                ('P6₃cm',[l, None, l]),
                                (u'P6\u0305c2',[l, None, l]),
                                ('P6₃/mcm',[l, None, l]),
                                ('P6cc',[l,l,l]),
                                ('P6/mcc',[l,l,l])
                              ])

    def getcolumn(self,m):
        mh,mk,ml = m
        mi = -(mh+mk)
        hexm = (mh,mk,mi,ml)
        if arezeros(1,1,1,0,hexm):
            column = 2
        elif mh==mk and mi==-2*mh:          # TODO: What rules apply to hexagonal 1111
            column = 1
        elif arezeros(0,0,1,0,hexm) and mh==-mk:
            column = 0
        else:
            column = None
        return column,m

class Cubic(SGClass):
    # Status: COMPLETE
    conditions = OrderedDict([('P23',[None, None, None, None]),
                                (u'Pm3\u0305',[None, None, None, None]),
                                ('P432',[None, None, None, None]),
                                (u'P4\u03053m',[None, None, None, None]),
                                (u'Pm3\u0305m',[None, None, None, None]),
                                ('P2₁3',[None, None, None, l]),
                                ('P4₂32',[None, None, None, l]),
                                ('P4₁32',[None, None, None, l4n]),
                                ('P4₃32',[None, None, None, l4n]),
                                (u'P4\u03053n',[None, None, l, l]),
                                (u'Pm3\u0305n',[None, None, l, l]),
                                (u'Pa3\u0305',[None, k, None, l]), # TODO: Handle this special case!
                                (u'Pn3\u0305',[None, kpl, None, l]),
                                (u'Pn3\u0305m',[None, kpl, None, l]),
                                (u'Pn3\u0305n',[None, kpl, l, l]),
                                ('I23', [hpkpl, kpl, l, l]),
                                ('I2₁3', [hpkpl, kpl, l, l]),
                                (u'Im3\u0305', [hpkpl, kpl, l, l]),
                                ('I432', [hpkpl, kpl, l, l]),
                                (u'I4\u03053m', [hpkpl, kpl, l, l]),
                                (u'Im3\u0305m (BCC)', [hpkpl, kpl, l, l]),
                                ('I4₁32',[hpkpl, kpl, l, l4n]),
                                (u'I4\u03053d',[hpkpl, kpl, [twohpl4n,l],l4n]),
                                (u'Ia3\u0305',[hpkpl, [k,l], l, l]),
                                (u'Ia3\u0305d',[hpkpl, [k,l], [twohpl4n,l],l4n]),
                                ('F23',[[hpk,hpl,kpl],[k,l],hpl,l]),
                                (u'Fm3\u0305',[[hpk,hpl,kpl],[k,l],hpl,l]),
                                ('F432',[[hpk,hpl,kpl],[k,l],hpl,l]),
                                (u'F4\u03053m',[[hpk,hpl,kpl],[k,l],hpl,l]),
                                (u'Fm3\u0305m',[[hpk,hpl,kpl],[k,l],hpl,l]),
                                ('F4₁32',[[hpk,hpl,kpl],[k,l],hpl,l4n]),
                                (u'F4\u03053c',[[hpk,hpl,kpl],[k,l],[h,l],l]),
                                (u'Fm3\u0305c',[[hpk,hpl,kpl],[k,l],[h,l],l]),
                                (u'Fd3\u0305',[[hpk,hpl,kpl],[kpl4n,k,l],hpl,l4n]),
                                (u'Fd3\u0305m',[[hpk,hpl,kpl],[kpl4n,k,l],hpl,l4n]),
                                (u'Fd3\u0305c',[[hpk,hpl,kpl],[kpl4n,k,l],[h,l],l4n])
                              ])

    def getcolumn(self,m):
        mh,mk,ml = m
        column = None
        for mh,mk,ml in [[mh,mk,ml],[mk,ml,mh],[ml,mh,mk]]:     # indicies are permutable for cubic (except #205)
            if arezeros(1,1,0,m):
                column = 3
            elif mh==mk:
                column = 2
            elif arezeros(1,0,0,m):
                column = 1

            if column:
                return column, [mh,mk,ml]
        column = 0
        return column, [mh,mk,ml]



Triclinic = Triclinic()
Monoclinic = Monoclinic()
Orthorhombic = Orthorhombic()
Tetragonal = Tetragonal()
Trigonal = Trigonal()
Hexagonal = Hexagonal()
Cubic = Cubic()


SGkeys = SGClass.conditions.keys()

crystalsystems = [Triclinic, Monoclinic, Orthorhombic,Tetragonal,Trigonal, Hexagonal, Cubic]


def check(m,SG):
    """
    This is the function intended for external use. Checks which crystal system the SG is in, then executes that
    system's check method.
    """

    # Check cache
    key=SG+','+''.join(map(str,m))
    if key in cache:
        return cache[key]

    for crystalsystem in crystalsystems:
        if SG in crystalsystem.conditions:
            reflected = crystalsystem.check(m,SG)
            cache[key]=reflected
            return reflected

if __name__=='__main__':
    def test(m,SG):
        print SG,m,check(m,SG)

    SG=u'P2₁'
    m=(1,2,1)
    test(m,SG)
    m=(0,1,0)
    test(m,SG)
    m=(0,2,0)
    test(m,SG)
    m=(0,3,0)
    test(m,SG)

    SG = u'P2₁2₁2'
    m=(1,1,1)
    test(m,SG)
    m=(0,1,0)
    test(m,SG)

    @debugtools.timeit
    def timetest():
        for i in range(0,300000):
            check(m,SG)
    timetest()

    m=(2,0,1)
    test(m,u'Pmn2₁')



# -*- coding: utf-8 -*-
import os
from time import sleep
from StringIO import StringIO
from PIL import Image
import numpy as np
from client.newt import NewtClient


class SpotClient(NewtClient):
    """
    Client class to handle SPOT API calls
    """

    BASE_DIR = '/global/project/projectdirs/als/spade/warehouse'
    SPOT_URL = 'https://portal-auth.nersc.gov/als'

    def __init__(self):
        super(SpotClient, self).__init__()
        self.spot_authentication = None

    def login(self, username, password):
        """
        Login to SPOT

        Parameters
        ----------
        username : str
        password : str
        """

        credentials = {"username": username,
                       "password": password}
        response = self.post(self.SPOT_URL + '/auth', data=credentials)
        if response.json()['auth']:
            self.authentication = response
            return super(SpotClient, self).login(username, password)
        else:
            self.authentication = None
            raise SPOTError('Bad Authentication: Unable to log in')

    def search(self, query, **kwargs):
        """
        Search a dataset on SPOT

        Parameters
        ----------
        query : str, search query
        kwargs : {key: option}
            Any of the following:
            'sortterm'
                attribute to sort results by
            'sorttype'
                ascending 'asc' or descending 'desc'
            'end_station'
                endstation of dataset
            'limitnum'
                maximum number of results to show
            'skipnum'
                number of results to skipd
            'search'
                search query

        Returns
        -------
        json
            response of search results
        """
        self.check_login()
        # sortterm probably allows more filters will need to check...
        allowed_kwargs = {'sortterm': ['fs.stage_date', 'appmetadata.sdate'],
                          'sorttype': ['desc', 'asc'],
                          'end_station': ['bl832', 'bl733']}
        generic_kwargs = ['limitnum', 'skipnum']
        for key in kwargs:
            if key not in allowed_kwargs.keys() and key not in generic_kwargs:
                raise ValueError('%s keyword not in allowed keywords %s' %
                                 (key, list(allowed_kwargs.keys() +
                                  generic_kwargs)))
            elif key in allowed_kwargs:
                if kwargs[key] not in allowed_kwargs[key]:
                    raise ValueError('%s keyword value must be on of %s' %
                                     (kwargs[key], list(allowed_kwargs[key])))
        kwargs.update(search=query)

        r = self.get(self.SPOT_URL + '/hdf/search', params=kwargs)

        return self.check_response(r)

    def get_derived_datasets(self, dataset):
        """
        Get datasets that are derived from the given dataset. ie raw, sino,
        norm, etc...

        Parameters
        ----------
        dataset : str
            dataset name

        Returns
        -------
        json
            Response with derived datasets
        """

        self.check_login()
        params = {'dataset': dataset}
        r = self.post(self.SPOT_URL + '/hdf/dataset', params=params)
        return self.check_response(r)

    def get_stage_path(self, dataset, stage):
        """
        Get the database path for a specific stage of a dataset. Stages
        are raw, norm, sino, gridrec, etc...

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name

        Returns
        -------
        str
            database path
        """
        self.check_login()
        derivatives = self.get_derived_datasets(dataset)

        for i in range(len(derivatives)):
            if derivatives[i]['stage'] == stage:
                path = derivatives[i]['path']
                return path

        raise SPOTError('Stage %s in dataset %s does not exist' %
                        (stage, dataset))

    def get_file_location(self, dataset, stage):
        """
        Get the full location (system path) for a specific stage of a dataset.
        stages are raw, norm, sino, gridrec, etc...

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name

        Returns
        -------
        str
            file path
        """
        self.check_login()
        derivatives = self.get_derived_datasets(dataset)

        for i in range(len(derivatives)):
            if derivatives[i]['stage'] == stage:
                location = derivatives[i]['phyloc']
                return location

        raise SPOTError('Stage %s in dataset %s does not exist' %
                        (stage, dataset))

    def get_dataset_attributes(self, dataset, stage='raw', group='/'):
        """
        Get hdf5 attributes of a specified dataset. group can be a specified
        image within the hdf5 file, or default '/' for top level attributes

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name
        group : str
            group name of hdf5

        Returns
        -------
        json
            json reponse of attributes
        """

        path = self.get_stage_path(dataset, stage)
        params = {'group': group}
        r = self.post(self.SPOT_URL + '/hdf/attributes' + path, params=params)

        return self.check_response(r)

    def list_dataset_images(self, dataset, stage):
        """
        List images inside a given dataset

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name

        Returns
        -------
        json
            response of image list
        """

        path = self.get_stage_path(dataset, stage)
        r = self.post(self.SPOT_URL + '/hdf/listimages' + path)

        return self.check_response(r)

    def get_dataset_size(self, dataset, stage):
        """
        Get the file size of a specified dataset

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name

        Returns
        -------
        int
            dataset size in bytes
        """

        path = self.get_stage_path(dataset, stage)
        r = self.session.head(self.SPOT_URL + '/hdf/download' + path)
        head = r.headers
        if not 'content-length' in head: return 1
        size = float(head['content-length'])

        return size

    def get_raw_image(self, dataset, stage, image=None, index=None):
        """
        Download raw data from an image in a SPOT dataset

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name
        image : str
            (optional), name of image in dataset
        index : int
            (optional) index of image in dataset (one of index or image must be given)
        Returns
        -------
        ndarray
            2D ndarray of image data
        """

        images = list(self.list_dataset_images(dataset, stage))
        if image is None and index is None:
            raise ValueError('One of image or index must be given')
        elif image is None and index is not None:
            group = images[index]
        else:
            group = os.path.split(images[0])[0] + '/' + image

        r = self.stage_tape_2_disk(dataset, stage)
        path = self.get_stage_path(dataset, stage)
        params = {'group': group}
        r = self.post(self.SPOT_URL + '/hdf/rawdata' + path, params=params)
        r = self.check_response(r)
        return np.array(r['data'])


    def get_image_download_URLS(self, dataset, stage, image=None, index=None):
        """
        Get download URL's for a specific image in a SPOT dataset

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name
        image : str
            (optional), name of image in dataset
        index : int
            (optional) index of image in dataset (one of index or image must be given)
        Returns
        -------
        dict
            Dictionary with urls to images
        """

        images = list(self.list_dataset_images(dataset, stage))
        if image is None and index is None:
            raise ValueError('One of image or index must be given')
        elif image is None and index is not None:
            group = images[index]
        else:
            group = os.path.split(images[0])[0] + '/' + image

        r = self.stage_tape_2_disk(dataset, stage)
        path = self.get_stage_path(dataset, stage)
        params = {'group': group}
        r = self.post(self.SPOT_URL + '/hdf/image' + path, params=params)
        r = self.check_response(r)
        return r

    def get_image_as(self, dataset, stage, ext='tif', image=None, index=None):
        """
        Download an image in the specified format and return an array of the image

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name
        ext : str, optional
            extension for image type (tif or png)
        image : str
            (optional), name of image in dataset
        index : int
            (optional) index of image in dataset (one of index or image must be given)
        Returns
        -------
        ndarray
            2D ndarray of image data
        """

        r = self.get_image_download_URLS(dataset, stage, image=image, index=index)
        url = r['pnglocaion'] if ext == 'png' else r['tiflocaion']  # Careful when spot API fixes this spelling mistake
        r = self.get(url)
        img = Image.open(StringIO(r.content))
        return np.asarray(img)


    def download_image(self, dataset, stage, save_path=None, ext='tif', image=None, index=None):
        """
        Download and save a specific image in a dataset as png or tif image

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name
        save_path : str, optional
            Path to save the image
        ext : str, optional
            extension for image type (tif or png)
        image : str
            (optional), name of image in dataset
        index : int
            (optional) index of image in dataset (one of index or image must be given)
        """

        if ext not in ('png', 'tif'):
            raise ValueError('ext can only be png or tif')
        if image is None and index is None:
            raise ValueError('One of image or index must be given')

        if save_path is None:
            name = image.split('.')[0] if image is not None else dataset + '_{}'.format(index)
            save_path = os.path.join(os.path.expanduser('~'), '{}.{}'.format(name,ext))

        r = self.get_image_download_URLS(dataset, stage, image=image, index=index)
        url = r['pnglocaion'] if ext == 'png' else r['tiflocaion'] # Careful when spot API fixes this spelling mistake
        r = self.get(url)

        with open(save_path, 'w') as f:
            for chunk in r:
                f.write(chunk)

    def stage_tape_2_disk(self, dataset, stage):
        """
        Stage a dataset from tape to disk if needed

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name
        """

        path = self.get_stage_path(dataset, stage)
        r = self.check_response(self.post(self.SPOT_URL + '/hdf/stageifneeded' + path))
        # Wait for staging to finish
        while r['location'] == 'unknown' or r['location'] == 'staging':
            sleep(3)
            r = self.check_response(self.post(self.SPOT_URL + '/hdf/stageifneeded' + path))

        return r

    def download_dataset(self, dataset, stage, save_path=None):
        """
        Download a specified dataset

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name
        save_path : str
            Path and name to save file locally. If None name on SPOT is used and save in home directory
        """

        path = self.get_stage_path(dataset, stage)
        if save_path is None:
            save_path = os.path.join(os.path.expanduser('~'), path.split('/')[-1])

        r = self.stage_tape_2_disk(dataset, stage)
        r = self.get(self.SPOT_URL + '/hdf/download' + path, stream=True)

        with open(save_path, 'wb') as f:
            for chunk in r.iter_content(chunk_size=64*1024):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    f.flush()
        r.close()

        return self.check_login(r)

    def download_dataset_generator(self, dataset, stage, save_path=None, chunk_size=64*1024):
        """
        Download a dataset as a generator (yields the fraction downloaded)
        Useful to know the status of a download (for gui purposes)

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name
        save_path : str
            path and name to save file locally. If None name on SPOT is used and save in home directory
        chunk_size : int
            Chuck size of data in bytes

        Yields
        ------
        float
            Percent downloaded
        """

        path = self.get_stage_path(dataset, stage)
        if save_path is None:
            save_path = os.path.join(os.path.expanduser('~'), path.split('/')[-1])

        r = self.stage_tape_2_disk(dataset, stage)
        file_size = float(self.get_dataset_size(dataset, stage))
        r = self.get(self.SPOT_URL + '/hdf/download' + path, stream=True)

        with open(save_path, 'wb') as f:
            downloaded = 0.0
            for chunk in r.iter_content(chunk_size=chunk_size):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    f.flush()
                    downloaded += len(chunk)/file_size
                    yield downloaded
        r.close()

        return

    def transfer_2_nersc(self, dataset, stage, path, system):
        # TODO need to find a way to make a generator out of this, here or from cp/rsync
        """
        Transfer a dataset to NERSC

        Parameters
        ----------
        dataset : str
            name of dataset
        stage : str
            stage name
        path : str
            absolute destination path on NERSC
        """

        r = self.stage_tape_2_disk(dataset, stage)
        if r['location'] != 'staging' or r['ocation'] != 'unknown':
            r = self.rsync(r['location'], path, system)

        return r

class SPOTError(Exception):
    """Raised when SPOT gets angry"""
    pass


if __name__ == '__main__':
    import time
    from StringIO import StringIO
    from PIL import Image
    from matplotlib.pyplot import imshow, show, figure
    s = SpotClient()
    s.login('lbluque', '')
    # t = time.time()
    # img = s.get_raw_image('20160630_054009_prefire_3_0amp_scan7', 'raw',  index=0)
    # print 'Time: ', time.time() - t
    t = time.time()
    arr = s.get_image_as('20160630_054009_prefire_3_0amp_scan7', 'raw', ext='tif', index=0)
    print arr.shape
    imshow(arr)
    show()
    # for i in range(3):
    #     figure(i)
    #     imshow(arr[:, :, i])
    # show()
    print 'Time: ', time.time() - t



from PySide.QtUiTools import QUiLoader
from PySide import QtGui
from PySide import QtCore

import models
import ui
import customwidgets
import display
import hig


features = []
functionTree = None
layout = None

def clearFeatures():
    global features
    while features:
        feature = features[-1]
        features.remove(feature)
        feature.deleteLater()


def addSubstrate():
    global features
    if not sum([type(feature) is customwidgets.substrate for feature in features]):
        features.insert(0, customwidgets.substrate())
    update()


def addLayer():
    global features
    features.append(customwidgets.layer())
    update()


def addParticle():
    global features
    features.append(customwidgets.particle())
    update()


def removeFeature(index):
    global features
    del features[index]
    update()


def layercount():
    return sum([type(feature) is customwidgets.layer for feature in features])


def particlecount():
    return sum([type(feature) is customwidgets.particle for feature in features])


def update():
    global functionTree
    assert isinstance(layout, QtGui.QVBoxLayout)

    for i in range(layout.count()):
        if layout.itemAt(i) not in features:
            layout.itemAt(i).parent = None

    # layout.addItem(QtGui.QSpacerItem(0,0,vData=QtGui.QSizePolicy.Expanding))

    for i, item in enumerate(features[::-1]):
        layout.insertWidget(i, item)

    if display.viewWidget:
        display.redraw()


def loadform(path):
    guiloader = QUiLoader()
    f = QtCore.QFile(path)
    f.open(QtCore.QFile.ReadOnly)
    form = guiloader.load(f)
    f.close()
    return form


def load():
    global features, functionTree
    layout.setAlignment(QtCore.Qt.AlignBottom)
    addSubstrate()
    addParticle()




import os
import string
from PySide import QtGui
import sys
import re
import msg
from appdirs import *

user_config_dir=user_config_dir('xicam')

def similarframe(path, N):
    """
    Get the file path N ahead (or behind) the provided path frame.
    """
    try:
        expr = '(?<=_)[\d]+(?=[_.])'
        frame = re.search(expr, os.path.basename(path)).group(0)
        leadingzeroslen=len(frame)
        framenum = int(frame)
        prevframenum = int(framenum) + N
        prevframenum = '{:0>{}}'.format(prevframenum,leadingzeroslen)
        return re.sub(expr, prevframenum, path)
    except ValueError:
        msg.logMessage('No earlier frame found for ' + path + ' with ' + N,msg.ERROR)
        return None


def path2nexus(path):
    """
    Get the path to corresponding nexus file
    """
    return os.path.splitext(path)[0] + '.nxs'


def getRoot():
    if sys.platform == 'linux2':
        return '/'
    elif sys.platform == 'darwin':
        return '/Volumes'
    elif sys.platform == 'win32':
        return QtGui.QFileSystemModel().myComputer()
    else:
        print 'WARNING: Unknown platform "' + sys.platform + '"'

    return None
